"""Unit tests for storage adapters."""

import pytest
import pandas as pd
import numpy as np
from pathlib import Path
import tempfile
import shutil

from meteaudata.storage.adapters.pandas_memory import PandasMemoryAdapter
from meteaudata.storage.adapters.pandas_disk import PandasDiskAdapter
from meteaudata.storage.config import StorageConfig
from meteaudata.storage.factory import create_backend

# Try to import optional backends
try:
    from meteaudata.storage.adapters.polars_adapter import PolarsAdapter
    import polars as pl

    POLARS_AVAILABLE = True
except ImportError:
    POLARS_AVAILABLE = False
    PolarsAdapter = None
    pl = None

try:
    from meteaudata.storage.adapters.sql_adapter import SQLAdapter

    SQLALCHEMY_AVAILABLE = True
except ImportError:
    SQLALCHEMY_AVAILABLE = False
    SQLAdapter = None


# Fixture for test data
@pytest.fixture
def sample_series():
    """Create a sample pandas Series for testing."""
    dates = pd.date_range("2024-01-01", periods=100, freq="h")
    values = np.random.randn(100)
    return pd.Series(values, index=dates, name="test_series")


@pytest.fixture
def sample_metadata():
    """Create sample metadata for testing."""
    return {
        "description": "Test time series",
        "units": "kg/m3",
        "processing_steps": [],
        "version": 1,
    }


# Tests for PandasMemoryAdapter
class TestPandasMemoryAdapter:
    """Tests for in-memory pandas storage adapter."""

    @pytest.fixture
    def adapter(self):
        """Create a PandasMemoryAdapter instance."""
        return PandasMemoryAdapter()

    def test_save_and_load(self, adapter, sample_series, sample_metadata):
        """Test saving and loading data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        loaded_series, loaded_metadata = adapter.load(key)

        pd.testing.assert_series_equal(loaded_series, sample_series)
        assert loaded_metadata == sample_metadata

    def test_exists(self, adapter, sample_series, sample_metadata):
        """Test checking if key exists."""
        key = "test_key"
        assert not adapter.exists(key)

        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

    def test_delete(self, adapter, sample_series, sample_metadata):
        """Test deleting data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

        adapter.delete(key)
        assert not adapter.exists(key)

    def test_delete_nonexistent_raises(self, adapter):
        """Test that deleting non-existent key raises KeyError."""
        with pytest.raises(KeyError):
            adapter.delete("nonexistent_key")

    def test_load_nonexistent_raises(self, adapter):
        """Test that loading non-existent key raises KeyError."""
        with pytest.raises(KeyError):
            adapter.load("nonexistent_key")

    def test_list_keys(self, adapter, sample_series, sample_metadata):
        """Test listing all keys."""
        assert adapter.list_keys() == []

        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        adapter.save(sample_series, "key3", sample_metadata)

        keys = adapter.list_keys()
        assert set(keys) == {"key1", "key2", "key3"}

    def test_clear(self, adapter, sample_series, sample_metadata):
        """Test clearing all data."""
        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        assert len(adapter) == 2

        adapter.clear()
        assert len(adapter) == 0
        assert adapter.list_keys() == []

    def test_to_pandas(self, adapter, sample_series):
        """Test to_pandas conversion (no-op for this adapter)."""
        result = adapter.to_pandas(sample_series)
        pd.testing.assert_series_equal(result, sample_series)

    def test_from_pandas(self, adapter, sample_series):
        """Test from_pandas conversion (no-op for this adapter)."""
        result = adapter.from_pandas(sample_series)
        pd.testing.assert_series_equal(result, sample_series)

    def test_backend_type(self, adapter):
        """Test backend type identifier."""
        assert adapter.get_backend_type() == "pandas-memory"

    def test_data_isolation(self, adapter, sample_series, sample_metadata):
        """Test that stored data is isolated from external modifications."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        # Modify original series
        sample_series.iloc[0] = 999.0

        # Load should return unmodified data
        loaded_series, _ = adapter.load(key)
        assert loaded_series.iloc[0] != 999.0

    def test_metadata_isolation(self, adapter, sample_series, sample_metadata):
        """Test that metadata is isolated from external modifications."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        # Modify original metadata
        sample_metadata["new_field"] = "new_value"

        # Load should not include new field
        _, loaded_metadata = adapter.load(key)
        assert "new_field" not in loaded_metadata


# Tests for PandasDiskAdapter
class TestPandasDiskAdapter:
    """Tests for on-disk pandas storage adapter."""

    @pytest.fixture
    def temp_dir(self):
        """Create a temporary directory for testing."""
        temp_path = Path(tempfile.mkdtemp())
        yield temp_path
        if temp_path.exists():
            shutil.rmtree(temp_path)

    @pytest.fixture
    def adapter(self, temp_dir):
        """Create a PandasDiskAdapter instance."""
        return PandasDiskAdapter(base_path=temp_dir)

    def test_save_and_load(self, adapter, sample_series, sample_metadata):
        """Test saving and loading data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        loaded_series, loaded_metadata = adapter.load(key)

        pd.testing.assert_series_equal(loaded_series, sample_series)
        assert loaded_metadata == sample_metadata

    def test_exists(self, adapter, sample_series, sample_metadata):
        """Test checking if key exists."""
        key = "test_key"
        assert not adapter.exists(key)

        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

    def test_delete(self, adapter, sample_series, sample_metadata):
        """Test deleting data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

        adapter.delete(key)
        assert not adapter.exists(key)

    def test_list_keys(self, adapter, sample_series, sample_metadata):
        """Test listing all keys."""
        assert adapter.list_keys() == []

        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        adapter.save(sample_series, "key3", sample_metadata)

        keys = adapter.list_keys()
        assert set(keys) == {"key1", "key2", "key3"}

    def test_clear(self, adapter, sample_series, sample_metadata):
        """Test clearing all data."""
        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        assert len(adapter) == 2

        adapter.clear()
        assert len(adapter) == 0

    def test_files_created(self, adapter, temp_dir, sample_series, sample_metadata):
        """Test that files are created on disk."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        data_path = adapter.get_data_path(key)
        meta_path = adapter.get_metadata_path(key)

        assert data_path.exists()
        assert meta_path.exists()

    def test_backend_type(self, adapter):
        """Test backend type identifier."""
        assert adapter.get_backend_type() == "pandas-disk"

    def test_get_size_on_disk(self, adapter, sample_series, sample_metadata):
        """Test getting storage size."""
        adapter.save(sample_series, "key1", sample_metadata)
        size = adapter.get_size_on_disk()
        assert size > 0

    def test_cleanup(self, adapter, temp_dir, sample_series, sample_metadata):
        """Test cleanup removes directory."""
        adapter.save(sample_series, "key1", sample_metadata)
        assert temp_dir.exists()

        adapter.cleanup()
        assert not temp_dir.exists()

    def test_persistence_across_instances(
        self, temp_dir, sample_series, sample_metadata
    ):
        """Test data persists across adapter instances."""
        # Save with first adapter
        adapter1 = PandasDiskAdapter(base_path=temp_dir)
        adapter1.save(sample_series, "test_key", sample_metadata)

        # Load with second adapter
        adapter2 = PandasDiskAdapter(base_path=temp_dir)
        loaded_series, loaded_metadata = adapter2.load("test_key")

        pd.testing.assert_series_equal(loaded_series, sample_series)
        assert loaded_metadata == sample_metadata


# Tests for PolarsAdapter (if available)
@pytest.mark.skipif(not POLARS_AVAILABLE, reason="Polars not installed")
class TestPolarsAdapter:
    """Tests for Polars storage adapter."""

    @pytest.fixture
    def temp_dir(self):
        """Create a temporary directory for testing."""
        temp_path = Path(tempfile.mkdtemp())
        yield temp_path
        if temp_path.exists():
            shutil.rmtree(temp_path)

    @pytest.fixture
    def adapter(self, temp_dir):
        """Create a PolarsAdapter instance."""
        return PolarsAdapter(base_path=temp_dir)

    def test_save_and_load_pandas(self, adapter, sample_series, sample_metadata):
        """Test saving pandas data and loading as Polars."""
        from meteaudata.types import IndexMetadata

        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        loaded_series, loaded_metadata = adapter.load(key)

        # Convert Polars back to pandas
        loaded_pandas = adapter.to_pandas(loaded_series)

        # Reconstruct index from stored values and metadata
        if "index_values" in loaded_metadata and "index_metadata" in loaded_metadata:
            # Create a pandas Index from the stored values
            index_values = loaded_metadata["index_values"]
            temp_index = pd.Index(index_values)

            # Apply metadata to reconstruct proper index type with properties
            index_meta = IndexMetadata(**loaded_metadata["index_metadata"])
            reconstructed_index = IndexMetadata.reconstruct_index(
                temp_index, index_meta
            )
            loaded_pandas.index = reconstructed_index

        pd.testing.assert_series_equal(loaded_pandas, sample_series)

    def test_save_and_load_polars(self, adapter, sample_series, sample_metadata):
        """Test saving and loading Polars data."""
        key = "test_key"
        polars_series = pl.from_pandas(sample_series)

        adapter.save(polars_series, key, sample_metadata)
        loaded_series, loaded_metadata = adapter.load(key)

        # Both should be Polars series
        assert isinstance(loaded_series, pl.Series)
        # Note: Polars conversion loses datetime index, so just compare values
        assert len(loaded_series) == len(polars_series)

    def test_to_pandas_conversion(self, adapter, sample_series):
        """Test conversion to pandas."""
        polars_series = pl.from_pandas(sample_series)
        result = adapter.to_pandas(polars_series)

        assert isinstance(result, pd.Series)
        # Note: Polars loses datetime index during conversion, so compare values
        assert len(result) == len(sample_series)
        np.testing.assert_array_almost_equal(result.values, sample_series.values)

    def test_from_pandas_conversion(self, adapter, sample_series):
        """Test conversion from pandas."""
        result = adapter.from_pandas(sample_series)

        assert isinstance(result, pl.Series)
        # Note: Polars loses datetime index, so just check values
        assert len(result) == len(sample_series)
        np.testing.assert_array_almost_equal(result.to_numpy(), sample_series.values)

    def test_backend_type(self, adapter):
        """Test backend type identifier."""
        assert adapter.get_backend_type() == "polars"


# Tests for SQLAdapter (if available)
@pytest.mark.skipif(not SQLALCHEMY_AVAILABLE, reason="SQLAlchemy not installed")
class TestSQLAdapter:
    """Tests for SQL database storage adapter."""

    @pytest.fixture
    def adapter(self):
        """Create a SQLAdapter instance with in-memory SQLite."""
        return SQLAdapter(connection_string="sqlite:///:memory:")

    def test_save_and_load(self, adapter, sample_series, sample_metadata):
        """Test saving and loading data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)

        loaded_series, loaded_metadata = adapter.load(key)

        # Note: Index might be converted to string, so compare values
        assert len(loaded_series) == len(sample_series)
        np.testing.assert_array_almost_equal(loaded_series.values, sample_series.values)
        assert loaded_metadata == sample_metadata

    def test_exists(self, adapter, sample_series, sample_metadata):
        """Test checking if key exists."""
        key = "test_key"
        assert not adapter.exists(key)

        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

    def test_delete(self, adapter, sample_series, sample_metadata):
        """Test deleting data."""
        key = "test_key"
        adapter.save(sample_series, key, sample_metadata)
        assert adapter.exists(key)

        adapter.delete(key)
        assert not adapter.exists(key)

    def test_list_keys(self, adapter, sample_series, sample_metadata):
        """Test listing all keys."""
        assert adapter.list_keys() == []

        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        adapter.save(sample_series, "key3", sample_metadata)

        keys = adapter.list_keys()
        assert set(keys) == {"key1", "key2", "key3"}

    def test_clear(self, adapter, sample_series, sample_metadata):
        """Test clearing all data."""
        adapter.save(sample_series, "key1", sample_metadata)
        adapter.save(sample_series, "key2", sample_metadata)
        assert len(adapter) == 2

        adapter.clear()
        assert len(adapter) == 0

    def test_execute_query(self, adapter, sample_series, sample_metadata):
        """Test executing custom SQL query."""
        adapter.save(sample_series, "test_key", sample_metadata)

        # Query metadata table
        results = adapter.execute_query(
            "SELECT key, series_name FROM time_series_metadata"
        )
        assert len(results) == 1
        assert results[0][0] == "test_key"

    def test_get_table_names(self, adapter):
        """Test getting table names."""
        tables = adapter.get_table_names()
        assert "time_series_data" in tables
        assert "time_series_metadata" in tables

    def test_backend_type(self, adapter):
        """Test backend type identifier."""
        assert adapter.get_backend_type() == "sql"

    def test_empty_series(self, adapter, sample_metadata):
        """Test saving and loading an empty series."""
        key = "empty_key"
        empty_series = pd.Series([], dtype=float, name="empty")

        adapter.save(empty_series, key, sample_metadata)
        loaded_series, _ = adapter.load(key)

        assert len(loaded_series) == 0


# Tests for factory and configuration
class TestStorageFactory:
    """Tests for storage backend factory."""

    def test_create_pandas_memory(self):
        """Test creating pandas memory backend."""
        config = StorageConfig.for_pandas_memory()
        backend = create_backend(config)

        assert isinstance(backend, PandasMemoryAdapter)
        assert backend.get_backend_type() == "pandas-memory"

    def test_create_pandas_disk(self):
        """Test creating pandas disk backend."""
        with tempfile.TemporaryDirectory() as temp_dir:
            config = StorageConfig.for_pandas_disk(temp_dir)
            backend = create_backend(config)

            assert isinstance(backend, PandasDiskAdapter)
            assert backend.get_backend_type() == "pandas-disk"

    @pytest.mark.skipif(not POLARS_AVAILABLE, reason="Polars not installed")
    def test_create_polars(self):
        """Test creating Polars backend."""
        with tempfile.TemporaryDirectory() as temp_dir:
            config = StorageConfig.for_polars(temp_dir)
            backend = create_backend(config)

            assert isinstance(backend, PolarsAdapter)
            assert backend.get_backend_type() == "polars"

    @pytest.mark.skipif(not SQLALCHEMY_AVAILABLE, reason="SQLAlchemy not installed")
    def test_create_sql(self):
        """Test creating SQL backend."""
        config = StorageConfig.for_sql("sqlite:///:memory:")
        backend = create_backend(config)

        assert isinstance(backend, SQLAdapter)
        assert backend.get_backend_type() == "sql"

    def test_create_sql_temp(self):
        """Test creating temporary SQL backend."""
        if not SQLALCHEMY_AVAILABLE:
            pytest.skip("SQLAlchemy not installed")

        config = StorageConfig.for_sql_temp()
        backend = create_backend(config)

        assert isinstance(backend, SQLAdapter)
        assert "sqlite:///" in backend.connection_string


class TestStorageConfig:
    """Tests for storage configuration."""

    def test_validate_pandas_disk_requires_path(self):
        """Test that pandas-disk requires base_path."""
        config = StorageConfig(backend_type="pandas-disk")
        with pytest.raises(ValueError, match="base_path is required"):
            config.validate_config()

    def test_validate_sql_requires_connection_string(self):
        """Test that sql requires connection_string."""
        config = StorageConfig(backend_type="sql")
        with pytest.raises(ValueError, match="connection_string is required"):
            config.validate_config()

    def test_pandas_memory_no_validation_error(self):
        """Test that pandas-memory validates without extra params."""
        config = StorageConfig(backend_type="pandas-memory")
        config.validate_config()  # Should not raise


class TestMetadataIdTracking:
    """Tests for metadata_id tracking in TimeSeries and Signal."""

    def test_timeseries_metadata_id_field_exists(self):
        """Test that TimeSeries has metadata_id field."""
        from meteaudata.types import TimeSeries
        import pandas as pd

        ts = TimeSeries(series=pd.Series([1, 2, 3], name="test"))
        assert hasattr(ts, "metadata_id")
        assert ts.metadata_id is None

    def test_timeseries_metadata_id_can_be_set(self):
        """Test that metadata_id can be set on TimeSeries."""
        from meteaudata.types import TimeSeries
        import pandas as pd

        ts = TimeSeries(series=pd.Series([1, 2, 3], name="test"))
        ts.metadata_id = "123"
        assert ts.metadata_id == "123"

    def test_timeseries_metadata_id_in_metadata_dict(self):
        """Test that metadata_id is included in metadata_dict()."""
        from meteaudata.types import TimeSeries
        import pandas as pd

        ts = TimeSeries(series=pd.Series([1, 2, 3], name="test"))
        ts.metadata_id = "456"

        metadata = ts.metadata_dict()
        assert "metadata_id" in metadata
        assert metadata["metadata_id"] == "456"

    def test_signal_provenance_metadata_id_propagates_to_raw_ts(self):
        """Test that Signal's provenance.metadata_id propagates to raw time series."""
        from meteaudata.types import Signal, TimeSeries
        import pandas as pd

        signal = Signal(
            name="test_signal",
            provenance={"metadata_id": "42", "parameter": "test"},
            input_data=pd.Series(
                [1, 2, 3], name="raw", index=pd.date_range("2024-01-01", periods=3)
            ),
        )

        # Check that raw time series has the metadata_id
        # Note: Signal name becomes "test_signal#1" and ts name is "test_signal#1_raw#1"
        raw_ts = signal.time_series.get("test_signal#1_raw#1")
        assert raw_ts is not None
        # Note: metadata_id is assigned during save_all(), not at creation

    def test_save_all_assigns_metadata_id_to_raw_ts(self):
        """Test that Signal.save_all() assigns metadata_id to raw time series."""
        from meteaudata.types import Signal
        from meteaudata.storage.adapters.pandas_memory import PandasMemoryAdapter
        import pandas as pd

        signal = Signal(
            name="test_signal",
            provenance={"metadata_id": "42", "parameter": "test"},
            input_data=pd.Series(
                [1, 2, 3], name="raw", index=pd.date_range("2024-01-01", periods=3)
            ),
        )

        adapter = PandasMemoryAdapter()
        signal._backend = adapter

        signal.save_all()

        # Note: Signal name becomes "test_signal#1" and ts name is "test_signal#1_raw#1"
        raw_ts = signal.time_series.get("test_signal#1_raw#1")
        assert raw_ts is not None
        assert raw_ts.metadata_id == "42"

    def test_save_all_allows_no_metadata_id_for_processed_ts_on_generic_backend(self):
        """Test that Signal.save_all() succeeds for processed TS with no metadata_id on generic backends.

        metadata_id is optional for backends that don't link to an external metadata registry
        (e.g. disk, memory, SQL). Only backends implementing generate_metadata_id() require it.
        """
        from meteaudata.types import Signal, TimeSeries
        from meteaudata.storage.adapters.pandas_memory import PandasMemoryAdapter
        import pandas as pd
        from datetime import datetime

        signal = Signal(
            name="test_signal",
            provenance={"metadata_id": "42", "parameter": "test"},
            input_data=pd.Series(
                [1, 2, 3], name="raw", index=pd.date_range("2024-01-01", periods=3)
            ),
        )

        # Add a processed time series without metadata_id
        processed_ts = TimeSeries(
            series=pd.Series(
                [4, 5, 6],
                name="processed",
                index=pd.date_range("2024-01-01", periods=3),
            )
        )
        signal.time_series["test_signal#1_processed#1"] = processed_ts

        adapter = PandasMemoryAdapter()
        signal._backend = adapter

        # Should succeed — metadata_id=None is valid for generic backends
        signal.save_all()
        processed = signal.time_series.get("test_signal#1_processed#1")
        assert processed is not None
        assert processed.metadata_id is None

    def test_save_all_uses_existing_metadata_id_on_processed_ts(self):
        """Test that Signal.save_all() uses existing metadata_id on processed TS."""
        from meteaudata.types import Signal, TimeSeries
        from meteaudata.storage.adapters.pandas_memory import PandasMemoryAdapter
        import pandas as pd

        signal = Signal(
            name="test_signal",
            provenance={"metadata_id": "42", "parameter": "test"},
            input_data=pd.Series(
                [1, 2, 3], name="raw", index=pd.date_range("2024-01-01", periods=3)
            ),
        )

        # Add a processed time series WITH metadata_id already set
        processed_ts = TimeSeries(
            series=pd.Series(
                [4, 5, 6],
                name="processed",
                index=pd.date_range("2024-01-01", periods=3),
            ),
            metadata_id="99",
        )
        signal.time_series["test_signal#1_processed#1"] = processed_ts

        adapter = PandasMemoryAdapter()
        signal._backend = adapter

        signal.save_all()

        processed = signal.time_series.get("test_signal#1_processed#1")
        assert processed is not None
        assert processed.metadata_id == "99"



import os
import re
from pathlib import Path
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

def insert_code_snippet(markdown_content: str, base_path: Path) -> str:
    """Insert code snippets into markdown content."""
    pattern = r"<!-- INSERT CODE: (.+?) -->"
    
    def replace_with_code(match):
        code_file_path = match.group(1).strip()
        
        # Handle different path formats
        if code_file_path.startswith('./docs/snippets/'):
            code_file_path = code_file_path.replace('./docs/snippets/', '')
        elif code_file_path.startswith('docs/snippets/'):
            code_file_path = code_file_path.replace('docs/snippets/', '')
        
        full_path = base_path / code_file_path
        try:
            with open(full_path, "r") as code_file:
                code_content = code_file.read()
            return f"```python\n{code_content}\n```"
        except FileNotFoundError:
            return f"<!-- ERROR: File not found: {code_file_path} at {full_path} -->"
    
    return re.sub(pattern, replace_with_code, markdown_content)

def build_readme():
    """Build the main README from template and snippets."""
    project_root = Path(__file__).parent.parent
    template_path = project_root / "docs" / "README_template.md"
    snippets_path = project_root / "docs" / "snippets"
    output_path = project_root / "README.md"
    
    print(f"Project root: {project_root}")
    print(f"Template path: {template_path}")
    print(f"Snippets path: {snippets_path}")
    print(f"Output path: {output_path}")
    
    if not template_path.exists():
        print(f"Template not found at {template_path}")
        return
    
    if not snippets_path.exists():
        print(f"Snippets directory not found at {snippets_path}")
        return
    
    with open(template_path, "r") as f:
        template_content = f.read()
    
    # Process the template to insert code snippets
    updated_content = insert_code_snippet(template_content, snippets_path)
    
    with open(output_path, "w") as f:
        f.write(updated_content)
    
    print(f"README.md generated at {output_path}")

class CustomBuildHook(BuildHookInterface):
    """Custom build hook to generate documentation."""
    
    def initialize(self, version, build_data):
        """Called before the build starts."""
        print("Generating documentation...")
        build_readme()
        print("Documentation generation complete.")

if __name__ == "__main__":
    build_readme()


from meteaudata.types import Dataset, Signal, TimeSeries
from test_metEAUdata import sample_dataset


def test_time_series_serde():
    dataset = sample_dataset()
    ts = dataset.signals["A#1"].time_series["A#1_RESAMPLED#1"]
    serialized = ts.model_dump_json()
    deserialised = TimeSeries.model_validate_json(serialized)
    assert ts == deserialised


def test_signal_serde():
    dataset = sample_dataset()
    signal = dataset.signals["A#1"]
    serialized = signal.model_dump_json()
    deserialised = Signal.model_validate_json(serialized)
    assert signal == deserialised


def test_dataset_serde():
    dataset = sample_dataset()
    serialized = dataset.model_dump_json()
    deserialised = Dataset.model_validate_json(serialized)
    assert dataset == deserialised



import datetime

import pandas as pd
from meteaudata.types import (
    FunctionInfo,
    Parameters,
    ProcessingStep,
    ProcessingType,
)


def subset(
    input_series: list[pd.Series],
    start_position,
    end_position,
    rank_based=False,
    *args,
    **kwargs,
) -> list[tuple[pd.Series, list[ProcessingStep]]]:
    """
    This function creates a subset of a time series.
    """
    func_info = FunctionInfo(
        name="subset",
        version="0.1",
        author="Jean-David Therrien",
        reference="www.github.com/modelEAU/meteaudata",
    )
    parameters = Parameters(
        start_position=start_position, end_position=end_position, rank_based=rank_based
    )
    processing_step = ProcessingStep(
        type=ProcessingType.RESAMPLING,
        parameters=parameters,
        function_info=func_info,
        description="A simple processing function that slices a series to given indices.",
        run_datetime=datetime.datetime.now(),
        requires_calibration=False,
        input_series_names=[str(col.name) for col in input_series],
        suffix="SLICE",
    )
    from meteaudata.types import Signal

    outputs = []
    for col in input_series:
        col = col.copy()
        col_name = col.name
        # Use utility function to handle new naming format (signalname_tsbase#number)
        signal, _, _ = Signal.extract_ts_base_and_number(str(col_name))
        if not isinstance(col.index, (pd.DatetimeIndex, pd.TimedeltaIndex)):
            raise IndexError(
                f"Series {col.name} has index type {type(col.index)}. Please provide either pd.DatetimeIndex or pd.TimedeltaIndex"
            )

        if rank_based:
            new_col = col.iloc[start_position:end_position]
        else:
            new_col = col.loc[start_position:end_position]

        new_name = f"{signal}_{processing_step.suffix}"
        new_col.name = new_name
        outputs.append((new_col, [processing_step]))
    return outputs



import datetime

import pandas as pd
from meteaudata.types import FunctionInfo, Parameters, ProcessingStep, ProcessingType

# this is a dummy value, replace it with the actual value if needed
some_argument = "dummy_argument"
some_value = "dummy_value"


def my_func(
    input_series: list[pd.Series], some_argument, some_keyword_argument=some_value
) -> list[tuple[pd.Series, list[ProcessingStep]]]:
    # Define the function information
    func_info = FunctionInfo(
        name="Double Values",
        version="1.0",
        author="Your Name",
        reference="www.yourwebsite.com",
    )

    # Define the processing step
    processing_step = ProcessingStep(
        type=ProcessingType.TRANSFORMATION,
        description="Doubles each value in the series",
        function_info=func_info,
        run_datetime=datetime.datetime.now(),
        requires_calibration=False,
        parameters=Parameters(
            some_argument=some_argument, some_keyword_argument=some_keyword_argument
        ),
        suffix="DBL",
    )
    # Example transformation logic
    outputs = []
    for series in input_series:
        transformed_series = series.apply(
            lambda x: x * 2
        )  # Example transformation: double the values

        # Append the transformed series and its processing steps
        outputs.append((transformed_series, [processing_step]))

    return outputs



import numpy as np
import pandas as pd
from meteaudata.processing_steps.multivariate.average import average_signals
from meteaudata.types import DataProvenance, Dataset, Signal

sample_data = np.random.randn(100, 3)
index = pd.date_range(start="2020-01-01", freq="6min", periods=100)

data = pd.DataFrame(sample_data, columns=["CODs", "NH4-N", "TSS"], index=index)
provenance_cods = DataProvenance(
    source_repository="metEAUdata README snippet",
    project="metEAUdata",
    location="Primary clarifier effluent",
    equipment="S::CAN Spectro::lyser no:xxxx",
    parameter="Soluble Chemical Oxygen Demand",
    purpose="Demonstrating how metEAUdata signals work",
    metadata_id="xyz",
)
signal_cods = Signal(
    input_data=data["CODs"].rename("RAW"),
    name="CODs",
    provenance=provenance_cods,
    units="mg/l",
)
provenance_nh4n = DataProvenance(
    source_repository="metEAUdata README snippet",
    project="metEAUdata",
    location="Primary clarifier effluent",
    equipment="S::CAN Ammo::lyser no:yyyy",
    parameter="Ammonium Nitrogen",
    purpose="Demonstrating how metEAUdata signals work",
    metadata_id="xyz",
)
signal_nh4n = Signal(
    input_data=data["NH4-N"].rename("RAW"),
    name="NH4-N",
    provenance=provenance_nh4n,
    units="mg/l",
)
# Create the Dataset
dataset = Dataset(
    name="test dataset",
    description="a small dataset with randomly generated data",
    owner="Jean-David Therrien",
    purpose="Demonstrating how metEAUdata datasets work",
    project="metEAUdata",
    signals={"CODs": signal_cods, "NH4-N": signal_nh4n},
)

# create a new signal by applying a transformation to items in the dataset
dataset.process(["CODs#1_RAW#1", "NH4-N#1_RAW#1"], average_signals)

print(dataset.signals["AVERAGE#1"])
# outputs Signal(name="AVERAGE#1", ...)
# The new signal has its own raw time series
print(dataset.signals["AVERAGE#1"].time_series["AVERAGE#1_RAW#1"])
# outputs TimeSeries(..., processing_steps=[<list of all the processing steps that went into creating CODs, NH4-N, and the averaged signal>])

# Save the resulting signal to a directory (data + metadata)
# dataset.save("test directory")

# Load a signal from a file
# dataset = Dataset.load(
#    "test directory/test dataset.zip",  # path to the dataset directory or zip file
#    "test dataset",  # name of the dataset
# )



"""In-memory pandas storage adapter.

This adapter maintains the current behavior of storing all data in memory
using pandas Series. It provides the baseline implementation and is the
default backend for backward compatibility.
"""

from typing import Any, Dict, List, Tuple
import pandas as pd
import copy


class PandasMemoryAdapter:
    """In-memory storage adapter using pandas Series.

    This adapter stores all data in a dictionary in memory, providing fast
    access but no memory optimization for large datasets. This is the default
    backend and maintains current system behavior.

    Attributes:
        _store: Dictionary mapping keys to (series, metadata) tuples
    """

    def __init__(self):
        """Initialize the in-memory storage."""
        self._store: Dict[str, Tuple[pd.Series, Dict[str, Any]]] = {}

    def save(self, data: pd.Series, key: str, metadata: Dict[str, Any]) -> None:
        """Save data with metadata to memory.

        Args:
            data: Pandas Series to store
            key: Unique identifier
            metadata: Metadata dictionary

        Note:
            Makes a copy of both data and metadata to prevent external modifications
        """
        # Make copies to prevent external modifications
        self._store[key] = (data.copy(), copy.deepcopy(metadata))

    def load(self, key: str) -> Tuple[pd.Series, Dict[str, Any]]:
        """Load data and metadata from memory.

        Args:
            key: Unique identifier

        Returns:
            Tuple of (series, metadata)

        Raises:
            KeyError: If key does not exist
        """
        if key not in self._store:
            raise KeyError(f"Key '{key}' not found in storage")

        series, metadata = self._store[key]
        # Return copies to prevent external modifications
        return series.copy(), copy.deepcopy(metadata)

    def delete(self, key: str) -> None:
        """Delete data from memory.

        Args:
            key: Unique identifier

        Raises:
            KeyError: If key does not exist
        """
        if key not in self._store:
            raise KeyError(f"Key '{key}' not found in storage")

        del self._store[key]

    def exists(self, key: str) -> bool:
        """Check if key exists in storage.

        Args:
            key: Unique identifier

        Returns:
            True if key exists, False otherwise
        """
        return key in self._store

    def list_keys(self) -> List[str]:
        """List all stored keys.

        Returns:
            List of all keys in storage
        """
        return list(self._store.keys())

    def to_pandas(self, data: pd.Series) -> pd.Series:
        """Convert to pandas Series (no-op for this adapter).

        Args:
            data: Pandas Series

        Returns:
            The same pandas Series
        """
        return data

    def from_pandas(self, series: pd.Series) -> pd.Series:
        """Convert from pandas Series (no-op for this adapter).

        Args:
            series: Pandas Series

        Returns:
            The same pandas Series
        """
        return series

    def clear(self) -> None:
        """Clear all stored data from memory.

        Warning:
            This operation is irreversible and removes all data.
        """
        self._store.clear()

    def get_backend_type(self) -> str:
        """Get backend type identifier.

        Returns:
            String 'pandas-memory'
        """
        return "pandas-memory"

    def __len__(self) -> int:
        """Get number of stored items.

        Returns:
            Number of keys in storage
        """
        return len(self._store)

    def __repr__(self) -> str:
        """String representation.

        Returns:
            Human-readable representation
        """
        return f"PandasMemoryAdapter(keys={len(self._store)})"



"""
Demo: cross-signal interactive dependency graph.

Three scenarios in one dataset:

  A) Divergence — RAW fans out into two resampled branches (5 min vs 10 min)
  B) Per-branch processing — each branch gets linear interpolation
  C) Cross-signal convergence — COD and NH4 interpolated series are averaged
     into a new AVERAGE signal via dataset.process()

Run:
    python demo_dependency_graph.py
"""
import numpy as np
import pandas as pd

from meteaudata.processing_steps.multivariate import average
from meteaudata.processing_steps.univariate import interpolate, replace, resample
from meteaudata.types import DataProvenance, Dataset, Signal

# ── Raw data ─────────────────────────────────────────────────────────────────

index = pd.date_range(start="2020-01-01", freq="6min", periods=200)
rng = np.random.default_rng(42)

cod_raw = pd.Series(rng.normal(50, 5, len(index)), index=index, name="RAW")
nh4_raw = pd.Series(rng.normal(10, 1, len(index)), index=index, name="RAW")

# Inject some gaps to make interpolation meaningful
cod_raw.iloc[30:40] = np.nan
nh4_raw.iloc[60:70] = np.nan

prov_cod = DataProvenance(
    source_repository="random",
    project="demo",
    location="reactor",
    equipment="probe",
    parameter="COD",
    purpose="dependency graph demo",
    metadata_id="cod-1",
)
prov_nh4 = DataProvenance(
    source_repository="random",
    project="demo",
    location="reactor",
    equipment="probe",
    parameter="NH4",
    purpose="dependency graph demo",
    metadata_id="nh4-1",
)

dataset = Dataset(
    name="demo",
    description="Dependency graph demo dataset",
    owner="demo",
    purpose="demo",
    project="demo",
    signals={
        "COD#1": Signal(input_data=cod_raw, name="COD#1", provenance=prov_cod, units="mg/L"),
        "NH4#1": Signal(input_data=nh4_raw, name="NH4#1", provenance=prov_nh4, units="mg/L"),
    },
)

# ── Scenario A: divergence — COD RAW → two resampled branches ────────────────

cod = dataset.signals["COD#1"]
cod = cod.process(["COD#1_RAW#1"], resample.resample, "5min",  output_names=["5MIN-RESAMPLED"])
cod = cod.process(["COD#1_RAW#1"], resample.resample, "10min", output_names=["10MIN-RESAMPLED"])
dataset.signals["COD#1"] = cod

# ── Scenario B: per-branch interpolation ─────────────────────────────────────

cod = cod.process(["COD#1_5MIN-RESAMPLED#1"],  interpolate.linear_interpolation, output_names=["LIN-INT-5MIN"])
cod = cod.process(["COD#1_10MIN-RESAMPLED#1"], interpolate.linear_interpolation, output_names=["LIN-INT-10MIN"])
dataset.signals["COD#1"] = cod

# NH4: simple single chain (for the convergence target)
nh4 = dataset.signals["NH4#1"]
nh4 = nh4.process(["NH4#1_RAW#1"], resample.resample, "5min")
nh4 = nh4.process(["NH4#1_RESAMPLED#1"], interpolate.linear_interpolation)
dataset.signals["NH4#1"] = nh4

# ── Scenario C: cross-signal convergence — average COD + NH4 ─────────────────

dataset = dataset.process(
    ["COD#1_LIN-INT-5MIN#1", "NH4#1_LIN-INT#1"],
    average.average_signals,
    check_units=False,
    output_signal_names=["AVG"],
)

# ── Summary ───────────────────────────────────────────────────────────────────

print("Signals in dataset:")
for sig_name, sig in dataset.signals.items():
    print(f"  {sig_name}: {list(sig.time_series.keys())}")

avg_ts_name = list(dataset.signals["AVG#1"].time_series.keys())[0]
print(f"\nAVERAGE time series name: {avg_ts_name}")

print("\n--- Scenario A+B: COD lineage for 5-min branch ---")
for dep in dataset.signals["COD#1"].build_dependency_graph("COD#1_LIN-INT-5MIN#1"):
    print(f"  {dep['origin']}  --[{dep['step']}]-->  {dep['destination']}")

print("\n--- Scenario C: full cross-signal lineage of the average ---")
for dep in dataset.build_dependency_graph(avg_ts_name):
    print(f"  {dep['origin']}  --[{dep['step']}]-->  {dep['destination']}")

# ── Interactive graphs ────────────────────────────────────────────────────────

print(f"\nOpening graph 1: lineage of {avg_ts_name} (cross-signal) ...")
dataset.plot_dependency_graph(avg_ts_name)

input("\nPress Enter to open graph 2: full dataset dependency graph ...")
dataset.plot_dependency_graph()



import numpy as np
import pytest

from meteaudata.types import (
    Parameters,
)  # Assuming your class is in a file called parameters.py


class TestParameters:
    def test_simple_parameters(self):
        """Test storing and retrieving simple parameters."""
        params = Parameters(a=1, b="string", c=3.14)
        assert params.a == 1
        assert params.b == "string"
        assert params.c == 3.14

    def test_as_dict_simple(self):
        """Test the as_dict method with simple types."""
        params = Parameters(a=1, b="string", c=3.14)
        data = params.as_dict()
        assert data == {"a": 1, "b": "string", "c": 3.14}

    def test_tuple_handling(self):
        """Test handling of tuple attributes."""
        simple_tuple = (1, 2, 3)
        nested_tuple = ((1, 2), (3, 4))
        mixed_tuple = (1, "string", 3.14)

        params = Parameters(
            simple_tuple=simple_tuple,
            nested_tuple=nested_tuple,
            mixed_tuple=mixed_tuple,
        )
        data = params.as_dict()

        assert data["simple_tuple"] == simple_tuple
        assert data["nested_tuple"] == nested_tuple
        assert data["mixed_tuple"] == mixed_tuple

        # Verify the types are preserved
        assert isinstance(data["simple_tuple"], tuple)
        assert isinstance(data["nested_tuple"], tuple)
        assert isinstance(data["mixed_tuple"], tuple)

    def test_dict_with_tuples(self):
        """Test handling of tuples inside dictionaries."""
        nested_data = {
            "level1": {
                "tuple": (5, 6, 7),
                "level2": {"tuple": ((8, 9), (10, 11))},
            }
        }

        params = Parameters(nested=nested_data)
        result = params.as_dict()

        # Check structure and tuple preservation
        assert isinstance(result["nested"]["level1"]["tuple"], tuple)
        assert isinstance(result["nested"]["level1"]["level2"]["tuple"], tuple)
        assert result["nested"]["level1"]["tuple"] == nested_data["level1"]["tuple"]
        assert (
            result["nested"]["level1"]["level2"]["tuple"]
            == nested_data["level1"]["level2"]["tuple"]
        )

    def test_numpy_array_handling(self):
        """Test handling of direct numpy arrays."""
        array_1d = np.array([1, 2, 3])
        array_2d = np.array([[1, 2], [3, 4]])

        params = Parameters(array_1d=array_1d, array_2d=array_2d)
        data = params.as_dict()

        # Check if as_dict() correctly restores the arrays
        assert isinstance(data["array_1d"], np.ndarray)
        assert isinstance(data["array_2d"], np.ndarray)
        assert np.array_equal(data["array_1d"], array_1d)
        assert np.array_equal(data["array_2d"], array_2d)

    def test_nested_dict_with_arrays(self):
        """Test handling of numpy arrays inside nested dictionaries."""
        nested_data = {
            "level1": {
                "array": np.array([5, 6, 7]),
                "level2": {"array": np.array([[8, 9], [10, 11]])},
            }
        }

        params = Parameters(nested=nested_data)
        result = params.as_dict()

        # Check structure and array restoration
        assert isinstance(result["nested"]["level1"]["array"], np.ndarray)
        assert isinstance(result["nested"]["level1"]["level2"]["array"], np.ndarray)
        assert np.array_equal(
            result["nested"]["level1"]["array"], nested_data["level1"]["array"]
        )
        assert np.array_equal(
            result["nested"]["level1"]["level2"]["array"],
            nested_data["level1"]["level2"]["array"],
        )

    def test_array_in_custom_object(self):
        """Test handling of custom objects containing numpy arrays."""

        class CustomObject:
            def __init__(self):
                self.array = np.array([12, 13, 14])
                self.name = "test_object"

        obj = CustomObject()
        params = Parameters(custom_obj=obj)
        result = params.as_dict()

        # Check that the custom object's array was handled correctly
        assert isinstance(result["custom_obj"], dict)  # Should be converted to dict
        assert isinstance(result["custom_obj"]["array"], np.ndarray)
        assert np.array_equal(result["custom_obj"]["array"], obj.array)
        assert result["custom_obj"]["name"] == "test_object"

    def test_list_of_arrays(self):
        """Test handling of lists containing numpy arrays."""
        arrays = [np.array([1, 2]), np.array([3, 4]), np.array([5, 6])]

        params = Parameters(array_list=arrays)
        result = params.as_dict()

        # Check each array in the list
        assert isinstance(result["array_list"], list)
        assert len(result["array_list"]) == 3
        for i in range(3):
            assert isinstance(result["array_list"][i], np.ndarray)
            assert np.array_equal(result["array_list"][i], arrays[i])

    def test_array_dtypes(self):
        """Test handling of arrays with different dtypes."""
        float_array = np.array([1.1, 2.2, 3.3], dtype=np.float64)
        int_array = np.array([1, 2, 3], dtype=np.int32)
        bool_array = np.array([True, False, True], dtype=bool)

        params = Parameters(
            float_array=float_array, int_array=int_array, bool_array=bool_array
        )
        result = params.as_dict()

        # Check dtype preservation
        assert result["float_array"].dtype == float_array.dtype
        assert result["int_array"].dtype == int_array.dtype
        assert result["bool_array"].dtype == bool_array.dtype

    def test_serialization_deserialization(self):
        """Test full serialization and deserialization cycle."""
        # Original parameters
        original = Parameters(
            name="test",
            array=np.array([1, 2, 3]),
            nested={"data": np.array([[4, 5], [6, 7]])},
        )

        # Serialize to dict
        serialized = original.model_dump()

        # Deserialize back to Parameters
        deserialized = Parameters.model_validate(serialized)
        result = deserialized.as_dict()

        # Verify everything was restored correctly
        assert result["name"] == "test"
        assert isinstance(result["array"], np.ndarray)
        assert np.array_equal(result["array"], np.array([1, 2, 3]))
        assert isinstance(result["nested"]["data"], np.ndarray)
        assert np.array_equal(result["nested"]["data"], np.array([[4, 5], [6, 7]]))

    def test_tuple_with_arrays(self):
        """Test handling of tuples containing numpy arrays."""
        array_tuple = (np.array([1, 2]), np.array([3, 4, 5]))

        params = Parameters(array_tuple=array_tuple)
        result = params.as_dict()

        # Check that the tuple structure is preserved but arrays are handled correctly
        assert isinstance(result["array_tuple"], tuple)
        assert len(result["array_tuple"]) == 2
        assert isinstance(result["array_tuple"][0], np.ndarray)
        assert isinstance(result["array_tuple"][1], np.ndarray)
        assert np.array_equal(result["array_tuple"][0], array_tuple[0])
        assert np.array_equal(result["array_tuple"][1], array_tuple[1])



"""Tests for export customization features added in v0.10.0"""
import os
import tempfile

import numpy as np
import pandas as pd
import pytest

from meteaudata.processing_steps.multivariate.average import average_signals
from meteaudata.processing_steps.univariate import resample
from meteaudata.types import DataProvenance, Dataset, Signal


def create_test_signal():
    """Create a simple signal for testing"""
    sample_data = pd.Series(
        np.random.randn(100),
        index=pd.date_range(start="2020-01-01", freq="1h", periods=100),
        name="RAW",
    )
    provenance = DataProvenance(
        source_repository="test",
        project="test project",
        location="test location",
        equipment="test equipment",
        parameter="temperature",
        purpose="testing",
        metadata_id="test_1",
    )
    signal = Signal(
        input_data=sample_data,
        name="A#1",
        provenance=provenance,
        units="°C",
    )
    return signal


def create_test_dataset():
    """Create a simple dataset for testing"""
    sample_data = pd.DataFrame(
        np.random.randn(100, 3),
        columns=["A", "B", "C"],
        index=pd.date_range(start="2020-01-01", freq="1h", periods=100),
    )
    provenance = DataProvenance(
        source_repository="test",
        project="test project",
        location="test location",
        equipment="test equipment",
        parameter="temperature",
        purpose="testing",
        metadata_id="test_1",
    )
    dataset = Dataset(
        name="test_dataset",
        description="Test dataset",
        owner="Test User",
        purpose="testing",
        project="test project",
        signals={
            "A#1": Signal(
                input_data=sample_data["A"].rename("RAW"),
                name="A#1",
                provenance=provenance,
                units="°C",
            ),
            "B#1": Signal(
                input_data=sample_data["B"].rename("RAW"),
                name="B#1",
                provenance=provenance,
                units="°C",
            ),
        },
    )
    return dataset


class TestCSVSeparator:
    """Test customizable CSV separator functionality"""

    def test_signal_save_with_comma_separator(self):
        """Test saving signal with default comma separator"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False, separator=",")
            # Check that CSV file exists
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            assert os.path.exists(csv_file)
            # Read the file and check separator
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert "," in first_line

    def test_signal_save_with_semicolon_separator(self):
        """Test saving signal with semicolon separator (European format)"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False, separator=";")
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            assert os.path.exists(csv_file)
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert ";" in first_line
                assert "," not in first_line.replace(";", "")

    def test_signal_save_with_tab_separator(self):
        """Test saving signal with tab separator"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False, separator="\t")
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            assert os.path.exists(csv_file)
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert "\t" in first_line

    def test_dataset_save_with_custom_separator(self):
        """Test saving dataset with custom separator propagates to all signals"""
        dataset = create_test_dataset()
        with tempfile.TemporaryDirectory() as temp_dir:
            dataset.save(temp_dir, separator=";")
            # Extract and check files
            import zipfile

            with zipfile.ZipFile(f"{temp_dir}/test_dataset.zip", "r") as zf:
                zf.extractall(f"{temp_dir}/extracted")

            # Check both signals have semicolon separator
            for signal_name in ["A#1", "B#1"]:
                csv_file = f"{temp_dir}/extracted/test_dataset_data/{signal_name}_data/{signal_name}_RAW#1.csv"
                assert os.path.exists(csv_file)
                with open(csv_file, "r") as f:
                    first_line = f.readline()
                    assert ";" in first_line


class TestIndexName:
    """Test customizable index name functionality"""

    def test_signal_save_with_custom_index_name(self):
        """Test saving signal with custom index name"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False, output_index_name="timestamp")
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            # Read CSV and check header
            df = pd.read_csv(csv_file, nrows=0)
            assert df.index.name is None or "timestamp" in df.columns or list(df.columns)[0] == "timestamp"
            # More robust check: read the first line
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert "timestamp" in first_line

    def test_signal_save_without_custom_index_name(self):
        """Test saving signal without custom index name uses default"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False)
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            assert os.path.exists(csv_file)

    def test_dataset_save_with_custom_index_name(self):
        """Test saving dataset with custom index name applies to all signals"""
        dataset = create_test_dataset()
        with tempfile.TemporaryDirectory() as temp_dir:
            dataset.save(temp_dir, output_index_name="datetime")
            # Extract and check files
            import zipfile

            with zipfile.ZipFile(f"{temp_dir}/test_dataset.zip", "r") as zf:
                zf.extractall(f"{temp_dir}/extracted")

            # Check both signals have custom index name
            for signal_name in ["A#1", "B#1"]:
                csv_file = f"{temp_dir}/extracted/test_dataset_data/{signal_name}_data/{signal_name}_RAW#1.csv"
                with open(csv_file, "r") as f:
                    first_line = f.readline()
                    assert "datetime" in first_line

    def test_combined_separator_and_index_name(self):
        """Test using both custom separator and index name together"""
        signal = create_test_signal()
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(temp_dir, zip=False, separator=";", output_index_name="time")
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_RAW#1.csv"
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert ";" in first_line
                assert "time" in first_line


class TestOutputNames:
    """Test custom output naming in process functions"""

    def test_signal_process_with_custom_output_name(self):
        """Test signal.process with custom output name"""
        signal = create_test_signal()
        signal = signal.process(
            ["A#1_RAW#1"],
            resample.resample,
            "1D",
            output_names=["daily-avg"],
        )
        # Check that the new time series has the custom name
        assert "A#1_daily-avg#1" in signal.all_time_series
        # Original should still exist
        assert "A#1_RAW#1" in signal.all_time_series

    def test_signal_process_custom_name_replaces_operation_suffix(self):
        """Test that custom name replaces the operation suffix, not signal name or hash"""
        signal = create_test_signal()
        signal = signal.process(
            ["A#1_RAW#1"],
            resample.resample,
            "6h",
            output_names=["sixhourly"],
        )
        # Should be A#1_sixhourly#1, not A#1_RESAMPLED#1
        assert "A#1_sixhourly#1" in signal.all_time_series
        assert "A#1_RESAMPLED#1" not in signal.all_time_series
        # Signal name should still be A#1
        assert signal.name == "A#1"

    def test_signal_process_multiple_outputs_with_custom_names(self):
        """Test signal.process with multiple outputs and custom names"""
        signal = create_test_signal()
        # Process twice to create multiple versions
        signal = signal.process(
            ["A#1_RAW#1"], resample.resample, "1D", output_names=["daily"]
        )
        signal = signal.process(
            ["A#1_RAW#1"], resample.resample, "1h", output_names=["hourly"]
        )
        assert "A#1_daily#1" in signal.all_time_series
        assert "A#1_hourly#1" in signal.all_time_series

    def test_signal_process_output_names_length_mismatch_raises_error(self):
        """Test that mismatched output_names length raises ValueError"""
        signal = create_test_signal()
        with pytest.raises(ValueError, match="output_names must have the same length"):
            # resample produces 1 output, but we provide 2 names
            signal.process(
                ["A#1_RAW#1"],
                resample.resample,
                "1D",
                output_names=["name1", "name2"],
            )

    def test_dataset_process_with_custom_output_name(self):
        """Test dataset.process with custom output signal name"""
        dataset = create_test_dataset()
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"],
            average_signals,
            output_signal_names=["siteaverage"],
        )
        # Check that the new signal has the custom name
        assert "siteaverage#1" in dataset.signals
        # Original signals should still exist
        assert "A#1" in dataset.signals
        assert "B#1" in dataset.signals


class TestOverwrite:
    """Test overwrite parameter in process functions"""

    def test_signal_process_without_overwrite_increments_hash(self):
        """Test that without overwrite, hash numbers increment"""
        signal = create_test_signal()
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")
        # Should have RESAMPLED#1 and RESAMPLED#2
        assert "A#1_RESAMPLED#1" in signal.all_time_series
        assert "A#1_RESAMPLED#2" in signal.all_time_series

    def test_signal_process_with_overwrite_keeps_hash(self):
        """Test that with overwrite=True, hash number doesn't increment"""
        signal = create_test_signal()
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")
        # Get the data from RESAMPLED#1
        first_result = signal.time_series["A#1_RESAMPLED#1"].series.copy()

        # Process again with overwrite=True
        signal = signal.process(
            ["A#1_RAW#1"], resample.resample, "1D", overwrite=True
        )

        # Should still only have RESAMPLED#1, no RESAMPLED#2
        assert "A#1_RESAMPLED#1" in signal.all_time_series
        assert "A#1_RESAMPLED#2" not in signal.all_time_series

    def test_signal_process_custom_name_with_overwrite(self):
        """Test combining custom output name with overwrite"""
        signal = create_test_signal()
        signal = signal.process(
            ["A#1_RAW#1"],
            resample.resample,
            "1D",
            output_names=["daily"],
        )
        assert "A#1_daily#1" in signal.all_time_series

        # Process again with overwrite=True and same custom name
        signal = signal.process(
            ["A#1_RAW#1"],
            resample.resample,
            "1D",
            output_names=["daily"],
            overwrite=True,
        )
        # Should still be daily#1, not daily#2
        assert "A#1_daily#1" in signal.all_time_series
        assert "A#1_daily#2" not in signal.all_time_series

    def test_signal_process_overwrite_uses_latest_number(self):
        """Test that overwrite=True overwrites the latest number, not #1"""
        signal = create_test_signal()
        # Create multiple versions
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")
        signal = signal.process(["A#1_RAW#1"], resample.resample, "1D")

        # Should have RESAMPLED#1, #2, and #3
        assert "A#1_RESAMPLED#1" in signal.all_time_series
        assert "A#1_RESAMPLED#2" in signal.all_time_series
        assert "A#1_RESAMPLED#3" in signal.all_time_series

        # Now process with overwrite=True - should overwrite #3, not #1
        signal = signal.process(
            ["A#1_RAW#1"], resample.resample, "1D", overwrite=True
        )

        # Should still have #1, #2, #3, but NOT #4
        assert "A#1_RESAMPLED#1" in signal.all_time_series
        assert "A#1_RESAMPLED#2" in signal.all_time_series
        assert "A#1_RESAMPLED#3" in signal.all_time_series
        assert "A#1_RESAMPLED#4" not in signal.all_time_series

    def test_dataset_process_without_overwrite_increments_hash(self):
        """Test that dataset.process increments hash without overwrite"""
        dataset = create_test_dataset()
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )
        # Should have AVERAGE#1 and AVERAGE#2
        assert "AVERAGE#1" in dataset.signals
        assert "AVERAGE#2" in dataset.signals

    def test_dataset_process_with_overwrite_keeps_hash(self):
        """Test that dataset.process with overwrite=True keeps hash"""
        dataset = create_test_dataset()
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals, overwrite=True
        )
        # Should still only have AVERAGE#1
        assert "AVERAGE#1" in dataset.signals
        assert "AVERAGE#2" not in dataset.signals

    def test_dataset_process_overwrite_uses_latest_number(self):
        """Test that overwrite=True overwrites the latest signal number, not #1"""
        dataset = create_test_dataset()
        # Create multiple versions
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals
        )

        # Should have AVERAGE#1, #2, and #3
        assert "AVERAGE#1" in dataset.signals
        assert "AVERAGE#2" in dataset.signals
        assert "AVERAGE#3" in dataset.signals

        # Now process with overwrite=True - should overwrite #3, not #1
        dataset = dataset.process(
            ["A#1_RAW#1", "B#1_RAW#1"], average_signals, overwrite=True
        )

        # Should still have #1, #2, #3, but NOT #4
        assert "AVERAGE#1" in dataset.signals
        assert "AVERAGE#2" in dataset.signals
        assert "AVERAGE#3" in dataset.signals
        assert "AVERAGE#4" not in dataset.signals


class TestIntegration:
    """Integration tests combining multiple features"""

    def test_full_workflow_with_all_features(self):
        """Test a complete workflow using all new features"""
        # Create a signal
        signal = create_test_signal()

        # Process with custom name and overwrite
        signal = signal.process(
            ["A#1_RAW#1"],
            resample.resample,
            "1D",
            output_names=["dailyaverage"],
            overwrite=False,
        )

        # Save with custom separator and index name
        with tempfile.TemporaryDirectory() as temp_dir:
            signal.save(
                temp_dir,
                zip=False,
                separator=";",
                output_index_name="date",
            )

            # Verify the saved file
            csv_file = f"{temp_dir}/{signal.name}_data/A#1_dailyaverage#1.csv"
            assert os.path.exists(csv_file)

            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert ";" in first_line
                assert "date" in first_line

    def test_dataset_full_workflow(self):
        """Test dataset workflow with all features"""
        dataset = create_test_dataset()

        # Process both signals
        for signal_name in ["A#1", "B#1"]:
            dataset.signals[signal_name] = dataset.signals[signal_name].process(
                [f"{signal_name}_RAW#1"],
                resample.resample,
                "1D",
                output_names=["daily"],
            )

        # Create an average signal with custom name
        dataset = dataset.process(
            ["A#1_daily#1", "B#1_daily#1"],
            average_signals,
            output_signal_names=["combinedaverage"],
            overwrite=False,
        )

        # Save with all custom options
        with tempfile.TemporaryDirectory() as temp_dir:
            dataset.save(temp_dir, separator="\t", output_index_name="timestamp")

            # Verify it was saved
            assert os.path.exists(f"{temp_dir}/test_dataset.zip")

            # Extract and spot-check one file
            import zipfile

            with zipfile.ZipFile(f"{temp_dir}/test_dataset.zip", "r") as zf:
                zf.extractall(f"{temp_dir}/extracted")

            csv_file = f"{temp_dir}/extracted/test_dataset_data/A#1_data/A#1_daily#1.csv"
            assert os.path.exists(csv_file)
            with open(csv_file, "r") as f:
                first_line = f.readline()
                assert "\t" in first_line
                assert "timestamp" in first_line



import numpy as np
import pandas as pd
import pytest

from meteaudata.processing_steps.multivariate import average
from meteaudata.processing_steps.univariate import (
    interpolate,
    prediction,
    replace,
    resample,
    subset,
)
from meteaudata.types import DataProvenance, Dataset, Signal, TimeSeries


def sample_dataset():
    sample_data = pd.DataFrame(
        np.random.randn(100, 3),
        columns=["A", "B", "C"],
        index=pd.date_range(start="2020-01-01", freq="6min", periods=100),
    )
    project = "PhD Thesis - metadata chapter"
    purpose = "Testing the metadata capture"
    provenance_a = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="COD",
        purpose=purpose,
        metadata_id="1",
    )
    provenance_b = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="NH4",
        purpose=purpose,
        metadata_id="2",
    )
    provenance_c = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="TSS",
        purpose=purpose,
        metadata_id="3",
    )
    dataset = Dataset(
        name="test dataset",
        description="a small dataset to test the metadata capture",
        owner="Jean-David Therrien",
        purpose=purpose,
        project=project,
        signals={
            "A#1": Signal(
                input_data=sample_data["A"].rename("RAW"),
                name="A#1",
                provenance=provenance_a,
                units="mg/l",
            ),
            "B#1": Signal(
                input_data=sample_data["B"].rename("RAW"),
                name="B#1",
                provenance=provenance_b,
                units="g/m3",
            ),
            "C#1": Signal(
                input_data=sample_data["C"].rename("RAW"),
                name="C#1",
                provenance=provenance_c,
                units="uS/cm",
            ),
            "D#1": Signal(
                input_data=sample_data["A"].rename("RAW"),
                name="D#1",
                provenance=provenance_a,
                units="mg/l",
            ),
        },
    )
    for signal_name, signal in dataset.signals.items():
        signal = signal.process([f"{signal_name}_RAW#1"], resample.resample, "5min")
        # introduce a nan in the resampled ts
        signal = signal.process(
            [f"{signal_name}_RESAMPLED#1"],
            replace.replace_ranges,
            [["2020-01-01 3:00:00", "2020-01-02 00:00:00"]],
            reason="sensor calibration procedure",
            replace_with=np.nan,
        )
        signal = signal.process(
            [f"{signal_name}_RESAMPLED#1"], interpolate.linear_interpolation
        )
    return dataset


def test_subset():
    dataset = sample_dataset()
    signal_names = dataset.all_signals
    signal_name = signal_names[0]
    signal = dataset.signals[signal_name]

    start = 2
    end = 7

    signal.process([f"{signal_name}_RESAMPLED#1"], subset.subset, start, end, True)
    assert len(signal.time_series[f"{signal_name}_SLICE#1"].series) == 5


def sample_dataset_no_nums():
    sample_data = pd.DataFrame(
        np.random.randn(100, 3),
        columns=["A", "B", "C"],
        index=pd.date_range(start="2020-01-01", freq="6min", periods=100),
    )
    project = "PhD Thesis - metadata chapter"
    purpose = "Testing the metadata capture"
    provenance_a = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="COD",
        purpose=purpose,
        metadata_id="1",
    )
    provenance_b = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="NH4",
        purpose=purpose,
        metadata_id="2",
    )
    provenance_c = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="TSS",
        purpose=purpose,
        metadata_id="3",
    )
    dataset = Dataset(
        name="test dataset",
        description="a small dataset to test the metadata capture",
        owner="Jean-David Therrien",
        purpose=purpose,
        project=project,
        signals={
            "A": Signal(
                input_data=sample_data["A"].rename("RAW"),
                name="A",
                provenance=provenance_a,
                units="mg/l",
            ),
            "B": Signal(
                input_data=sample_data["B"].rename("RAW"),
                name="B",
                provenance=provenance_b,
                units="g/m3",
            ),
            "C": Signal(
                input_data=sample_data["C"].rename("RAW"),
                name="C",
                provenance=provenance_c,
                units="uS/cm",
            ),
            "D": Signal(
                input_data=TimeSeries(series=sample_data["A"].rename("RAW")),
                name="D",
                provenance=provenance_a,
                units="mg/l",
            ),
        },
    )
    for signal_name, signal in dataset.signals.items():
        signal = signal.process([f"{signal_name}_RAW#1"], resample.resample, "5min")
        # introduce a nan in the resampled ts
        signal = signal.process(
            [f"{signal_name}_RESAMPLED#1"],
            replace.replace_ranges,
            [("2020-01-01 3:00:00", "2020-01-02 00:00:00")],
            reason="sensor calibration procedure",
            replace_with=np.nan,
        )
        signal = signal.process(
            [f"{signal_name}_REPLACED-RANGES#1"], interpolate.linear_interpolation
        )

    return dataset


def test_save_reread() -> None:
    dataset = sample_dataset()
    dataset.save("./tests/metadeauta_out")
    dataset2 = Dataset.load("./tests/metadeauta_out/test dataset.zip", dataset.name)
    # inspect every attribute of the dataset and see if they match
    for signal_name, signal in dataset.signals.items():
        signal2 = dataset2.signals[signal_name]
        assert signal.name == signal2.name
        assert signal.units == signal2.units
        assert signal.provenance == signal2.provenance
        assert signal.time_series.keys() == signal2.time_series.keys()
        for ts_name, ts in signal.time_series.items():
            print("series are equal?", ts_name)
            ts2 = signal2.time_series[ts_name]
            assert np.allclose(ts.series.values, ts2.series.values, equal_nan=True)  # type: ignore
            assert ts.processing_steps == ts2.processing_steps
            assert ts.index_metadata == ts2.index_metadata

    assert dataset == dataset2


def test_plots():
    dataset = sample_dataset()
    # add a prediction step to the dataset
    dataset.signals["A#1"] = dataset.signals["A#1"].process(
        ["A#1_LIN-INT#1"], prediction.predict_from_previous_point
    )
    fig = dataset.signals["A#1"].time_series["A#1_PREV-PRED#1"].plot()
    assert fig is not None
    fig = dataset.signals["A#1"].plot(
        ts_names=["A#1_RAW#1", "A#1_RESAMPLED#1", "A#1_LIN-INT#1", "A#1_PREV-PRED#1"],
        title="Sample graph",
    )
    assert fig is not None
    fig = dataset.plot(
        signal_names=["A#1", "B#1", "C#1"],
        ts_names=[
            "A#1_RAW#1",
            "A#1_RESAMPLED#1",
            "B#1_LIN-INT#1",
            "B#1_PREV-PRED#1",
            "C#1_RAW#1",
            "C#1_RESAMPLED#1",
            "C#1_LIN-INT#1",
            "C#1_PREV-PRED#1",
        ],
        title="Sample graph",
    )
    assert fig is not None
    from unittest.mock import patch
    with patch("webbrowser.open"):
        html_path = dataset.signals["A#1"].plot_dependency_graph("A#1_PREV-PRED#1")
        assert isinstance(html_path, str)
        html_path = dataset.signals["A#1"].plot_dependency_graph("A#1_RAW#1")
        assert isinstance(html_path, str)


def test_processing():
    dataset = sample_dataset()
    signal = dataset.signals["A#1"]
    # add a prediction step to the dataset
    signal = signal.process(["A#1_LIN-INT#1"], prediction.predict_from_previous_point)
    for ts_name in signal.all_time_series:
        ts = signal.time_series[ts_name]
        steps = ts.processing_steps
        for step in steps:
            input_series_names = step.input_series_names
            for input_series_name in input_series_names:
                assert input_series_name in signal.all_time_series


def test_dataset_unnumbered_signals():
    dataset = sample_dataset_no_nums()
    assert "A#1" in dataset.signals.keys()
    assert "A#1_RAW#1" in dataset.signals["A#1"].time_series


def test_dateset_numbered_signals():
    sample_data = pd.DataFrame(
        np.random.randn(100, 3),
        columns=["A", "B", "C"],
        index=pd.date_range(start="2020-01-01", freq="6min", periods=100),
    )
    project = "PhD Thesis - metadata chapter"
    purpose = "Testing the metadata capture"
    provenance_a = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="COD",
        purpose=purpose,
        metadata_id="1",
    )
    provenance_b = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="NH4",
        purpose=purpose,
        metadata_id="2",
    )
    provenance_c = DataProvenance(
        source_repository="random generation",
        project=project,
        location="CPU",
        equipment="numpy",
        parameter="TSS",
        purpose=purpose,
        metadata_id="3",
    )
    dataset = Dataset(
        name="test dataset",
        description="a small dataset to test the metadata capture",
        owner="Jean-David Therrien",
        purpose=purpose,
        project=project,
        signals={
            "A#1": Signal(
                input_data=sample_data["A"].rename("RAW#1"),
                name="A#1",
                provenance=provenance_a,
                units="mg/l",
            ),
            "B#1": Signal(
                input_data=sample_data["B"].rename("RAW#1"),
                name="B#1",
                provenance=provenance_b,
                units="g/m3",
            ),
            "C#1": Signal(
                input_data=sample_data["C"].rename("RAW#1"),
                name="C#1",
                provenance=provenance_c,
                units="uS/cm",
            ),
        },
    )
    assert "A#1" in dataset.signals.keys()
    assert "A#1_RAW#1" in dataset.signals["A#1"].time_series


def test_multivariate_average():
    dataset = sample_dataset()

    # assert that this raises a ValueError
    with pytest.raises(ValueError):
        dataset.process(
            ["A#1_RESAMPLED#1", "B#1_RESAMPLED#1", "C#1_RESAMPLED#1"],
            average.average_signals,
        )
    dataset.signals["B#1"].units = "mg/l"
    dataset.signals["C#1"].units = "mg/l"

    dataset = dataset.process(
        ["A#1_RESAMPLED#1", "B#1_RESAMPLED#1", "C#1_RESAMPLED#1"],
        average.average_signals,
    )
    assert "AVERAGE#1" in dataset.signals
    assert dataset.signals["AVERAGE#1"].units == "mg/l"
    assert dataset.signals["AVERAGE#1"].provenance == dataset.signals["A#1"].provenance
    assert len(dataset.signals["AVERAGE#1"].time_series) == 1
    assert "AVERAGE#1_RAW#1" in dataset.signals["AVERAGE#1"].time_series
    assert (
        len(
            dataset.signals["AVERAGE#1"].time_series["AVERAGE#1_RAW#1"].processing_steps
        )
        == len(dataset.signals["A#1"].time_series["A#1_RESAMPLED#1"].processing_steps)
        + len(dataset.signals["B#1"].time_series["B#1_RESAMPLED#1"].processing_steps)
        + +len(dataset.signals["C#1"].time_series["C#1_RESAMPLED#1"].processing_steps)
        + 1
    )


if __name__ == "__main__":
    test_save_reread()



"""On-disk pandas storage adapter using Parquet files.

This adapter stores data on disk using the efficient Parquet format,
enabling processing of datasets larger than available memory.
"""

from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional
import pandas as pd
import yaml
import tempfile
import shutil

from meteaudata.types import IndexMetadata


class PandasDiskAdapter:
    """On-disk storage adapter using pandas and Parquet format.

    This adapter stores time series data as Parquet files and metadata as
    YAML files in a specified directory. Data is only loaded into memory
    when accessed, enabling memory-efficient processing of large datasets.

    Attributes:
        base_path: Base directory for storing data files
    """

    def __init__(self, base_path: Optional[Path] = None):
        """Initialize the on-disk storage adapter.

        Args:
            base_path: Directory to store files. If None, creates a temporary directory.
        """
        if base_path is None:
            base_path = Path(tempfile.mkdtemp(prefix="meteaudata_"))

        self._base_path = Path(base_path)
        self._base_path.mkdir(parents=True, exist_ok=True)

    @property
    def base_path(self) -> Path:
        """Get the base path for file storage.

        Returns:
            Path to the base directory
        """
        return self._base_path

    def get_data_path(self, key: str) -> Path:
        """Get the file path for storing data.

        Args:
            key: Unique identifier

        Returns:
            Path to the Parquet data file
        """
        # Sanitize key for filesystem
        safe_key = key.replace("/", "_").replace("\\", "_")
        return self._base_path / f"{safe_key}.parquet"

    def get_metadata_path(self, key: str) -> Path:
        """Get the file path for storing metadata.

        Args:
            key: Unique identifier

        Returns:
            Path to the YAML metadata file
        """
        safe_key = key.replace("/", "_").replace("\\", "_")
        return self._base_path / f"{safe_key}_meta.yaml"

    def save(self, data: pd.Series, key: str, metadata: Dict[str, Any]) -> None:
        """Save data and metadata to disk.

        Args:
            data: Pandas Series to store
            key: Unique identifier
            metadata: Metadata dictionary
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        # Extract and store index metadata
        index_metadata = IndexMetadata.extract_index_metadata(data.index)
        metadata['index_metadata'] = index_metadata.model_dump()

        # Convert Series to DataFrame for Parquet storage
        # Preserve series name and index
        df = data.to_frame()

        # Save data as Parquet
        df.to_parquet(data_path, engine='pyarrow', compression='snappy')

        # Save metadata as YAML
        with open(meta_path, 'w') as f:
            yaml.dump(metadata, f, default_flow_style=False, allow_unicode=True)

    def load(self, key: str) -> Tuple[pd.Series, Dict[str, Any]]:
        """Load data and metadata from disk.

        Args:
            key: Unique identifier

        Returns:
            Tuple of (series, metadata)

        Raises:
            KeyError: If key does not exist
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        if not data_path.exists():
            raise KeyError(f"Key '{key}' not found in storage (path: {data_path})")

        # Load data from Parquet
        df = pd.read_parquet(data_path, engine='pyarrow')

        # Convert back to Series
        series = df.iloc[:, 0]

        # Load metadata from YAML
        if meta_path.exists():
            with open(meta_path, 'r') as f:
                metadata = yaml.safe_load(f) or {}
        else:
            metadata = {}

        # Reconstruct index with metadata if available
        if 'index_metadata' in metadata:
            index_meta = IndexMetadata(**metadata['index_metadata'])
            reconstructed_index = IndexMetadata.reconstruct_index(
                series.index,
                index_meta
            )
            series.index = reconstructed_index

        return series, metadata

    def delete(self, key: str) -> None:
        """Delete data and metadata files from disk.

        Args:
            key: Unique identifier

        Raises:
            KeyError: If key does not exist
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        if not data_path.exists():
            raise KeyError(f"Key '{key}' not found in storage")

        # Delete both data and metadata files
        data_path.unlink()
        if meta_path.exists():
            meta_path.unlink()

    def exists(self, key: str) -> bool:
        """Check if key exists in storage.

        Args:
            key: Unique identifier

        Returns:
            True if data file exists, False otherwise
        """
        return self.get_data_path(key).exists()

    def list_keys(self) -> List[str]:
        """List all stored keys.

        Returns:
            List of all keys in storage
        """
        keys = []
        for parquet_file in self._base_path.glob("*.parquet"):
            # Extract key from filename (remove .parquet extension)
            key = parquet_file.stem
            keys.append(key)
        return sorted(keys)

    def to_pandas(self, data: pd.Series) -> pd.Series:
        """Convert to pandas Series (no-op for this adapter).

        Args:
            data: Pandas Series

        Returns:
            The same pandas Series
        """
        return data

    def from_pandas(self, series: pd.Series) -> pd.Series:
        """Convert from pandas Series (no-op for this adapter).

        Args:
            series: Pandas Series

        Returns:
            The same pandas Series
        """
        return series

    def clear(self) -> None:
        """Clear all stored data from disk.

        Warning:
            This operation is irreversible and removes all files.
        """
        for file_path in self._base_path.glob("*.parquet"):
            file_path.unlink()
        for file_path in self._base_path.glob("*.yaml"):
            file_path.unlink()

    def cleanup(self) -> None:
        """Remove the entire storage directory.

        Warning:
            This is a destructive operation that removes the base directory
            and all its contents. Use with caution.
        """
        if self._base_path.exists():
            shutil.rmtree(self._base_path)

    def get_backend_type(self) -> str:
        """Get backend type identifier.

        Returns:
            String 'pandas-disk'
        """
        return "pandas-disk"

    def get_size_on_disk(self) -> int:
        """Get total size of stored data in bytes.

        Returns:
            Total size in bytes
        """
        total_size = 0
        for file_path in self._base_path.glob("*"):
            if file_path.is_file():
                total_size += file_path.stat().st_size
        return total_size

    def __len__(self) -> int:
        """Get number of stored items.

        Returns:
            Number of keys in storage
        """
        return len(self.list_keys())

    def __repr__(self) -> str:
        """String representation.

        Returns:
            Human-readable representation
        """
        return f"PandasDiskAdapter(base_path='{self._base_path}', keys={len(self)})"



import tempfile

import pandas as pd
import pytest
from meteaudata.types import (
    IndexMetadata,
)


def test_extract_datetime_index_metadata():
    # Create a sample DateTimeIndex
    dt_index = pd.date_range(
        start="2020-01-01", periods=3, freq="D", name="date_index", tz="UTC"
    )

    # Extract metadata using the method
    metadata = IndexMetadata.extract_index_metadata(dt_index)

    # Assertions to verify the extracted metadata
    assert metadata.type == "DatetimeIndex"
    assert metadata.name == "date_index"
    assert metadata.frequency == "D"
    assert metadata.time_zone == "UTC"


def test_reconstruct_datetime_index():
    # Create a sample DateTimeIndex
    dt_index = pd.date_range(start="2020-01-01", periods=3, freq="D")
    values = [1, 2, 3]
    series = pd.Series(values, index=dt_index)
    # simulate saving the index to a csv and reloading it using read_csv
    # save to a temporary file
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = temp_dir + "/temp.csv"
        series.to_csv(temp_file)
        # read the file back
        recovered_series = pd.read_csv(temp_file, parse_dates=True, index_col=0)

    # Create metadata manually or extract from another index as needed
    metadata = IndexMetadata(
        type="DatetimeIndex",
        dtype="datetime64[ns]",
        name="date_index",
        frequency="D",
        time_zone=None,
    )

    # Reconstruct the index from metadata
    reconstructed_index = IndexMetadata.reconstruct_index(
        recovered_series.index, metadata
    )

    # Assertions to verify the reconstructed index
    assert isinstance(reconstructed_index, pd.DatetimeIndex)
    assert reconstructed_index.name == "date_index"
    assert reconstructed_index.freqstr == "D"
    assert str(reconstructed_index.tz) == "None"


def test_reconstruct_datetime_index_tz():
    # Create a sample DateTimeIndex
    dt_index = pd.date_range(start="2020-01-01", periods=3, freq="D", tz="UTC")
    values = [1, 2, 3]
    series = pd.Series(values, index=dt_index)
    # simulate saving the index to a csv and reloading it using read_csv
    # save to a temporary file
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = temp_dir + "/temp.csv"
        series.to_csv(temp_file)
        # read the file back
        recovered_series = pd.read_csv(temp_file, parse_dates=True, index_col=0)

    # Create metadata manually or extract from another index as needed
    metadata = IndexMetadata(
        type="DatetimeIndex",
        dtype="datetime64[ns]",
        name="date_index",
        frequency="D",
        time_zone="UTC",
    )

    # Reconstruct the index from metadata
    reconstructed_index = IndexMetadata.reconstruct_index(
        recovered_series.index, metadata
    )

    # Assertions to verify the reconstructed index
    assert isinstance(reconstructed_index, pd.DatetimeIndex)
    assert reconstructed_index.name == "date_index"
    assert reconstructed_index.freqstr == "D"
    assert str(reconstructed_index.tz) == "UTC"


@pytest.mark.parametrize(
    "index_type, create_index, metadata_attrs",
    [
        (
            "DatetimeIndex",
            lambda: pd.date_range(start="2020-01-01", periods=3, freq="D", tz="UTC"),
            {
                "type": "DatetimeIndex",
                "dtype": "datetime64[ns, UTC]",
                "name": "date_index",
                "frequency": "D",
                "time_zone": "UTC",
            },
        ),
        (
            "DatetimeIndex",
            lambda: pd.date_range(start="2020-01-01", periods=3, freq="D"),
            {
                "type": "DatetimeIndex",
                "dtype": "datetime64[ns]",
                "name": "date_index",
                "frequency": "D",
                "time_zone": None,
            },
        ),
        (
            "Float64Index",
            lambda: pd.Index([0.1, 0.2, 0.3], dtype="float64"),
            {"type": "Index", "dtype": "float64", "name": "float_index"},
        ),
        (
            "Int64Index",
            lambda: pd.Index([1, 2, 3], dtype="int64"),
            {"type": "Index", "dtype": "int64", "name": "int_index"},
        ),
        (
            "RangeIndex",
            lambda: pd.RangeIndex(start=0, stop=3, step=1),
            {
                "type": "RangeIndex",
                "dtype": "int64",
                "name": "range_index",
                "start": 0,
                "end": 3,
                "step": 1,
            },
        ),
        (
            "CategoricalIndex",
            lambda: pd.CategoricalIndex(["a", "b", "c"]),
            {
                "type": "CategoricalIndex",
                "dtype": "category",
                "name": "cat_index",
                "categories": ["a", "b", "c"],
                "ordered": False,
            },
        ),
        (
            "PeriodIndex",
            lambda: pd.period_range(start="2020-01-01", periods=3, freq="M"),
            {
                "type": "PeriodIndex",
                "dtype": "period[M]",
                "name": "period_index",
                "frequency": "M",
            },
        ),
    ],
)
def test_reconstruct_index(index_type, create_index, metadata_attrs):
    original_index = create_index()
    original_index.name = metadata_attrs["name"]
    values = range(len(original_index))
    series = pd.Series(values, index=original_index)
    parse_dates = True if index_type == "DatetimeIndex" else False
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = temp_dir + "/temp.csv"
        series.to_csv(temp_file)
        recovered_series = pd.read_csv(temp_file, parse_dates=parse_dates, index_col=0)

    expected_metadata = IndexMetadata(**metadata_attrs)
    extracted_metadata_original_index = IndexMetadata.extract_index_metadata(
        original_index
    )
    for attr in metadata_attrs:
        assert getattr(expected_metadata, attr) == getattr(
            extracted_metadata_original_index, attr
        )

    reconstructed_index = IndexMetadata.reconstruct_index(
        recovered_series.index, expected_metadata
    )
    extracted_metadata_reconstructed_index = IndexMetadata.extract_index_metadata(
        reconstructed_index
    )
    for attr in metadata_attrs:
        assert getattr(expected_metadata, attr) == getattr(
            extracted_metadata_reconstructed_index, attr
        )

    if index_type == "DatetimeIndex":
        if metadata_attrs["time_zone"] is not None:
            assert str(reconstructed_index.tz) == metadata_attrs["time_zone"]
        assert reconstructed_index.freqstr == metadata_attrs["frequency"]

    elif index_type == "RangeIndex":
        assert reconstructed_index.start == metadata_attrs["start"]
        assert reconstructed_index.stop == metadata_attrs["end"]
        assert reconstructed_index.step == metadata_attrs["step"]
    elif index_type == "CategoricalIndex":
        assert reconstructed_index.categories.tolist() == metadata_attrs["categories"]
        assert reconstructed_index.ordered == metadata_attrs["ordered"]
    elif index_type == "PeriodIndex":
        assert reconstructed_index.freqstr == metadata_attrs["frequency"]
    else:
        pass
    return



"""Integration tests for storage backend support in TimeSeries, Signal, and Dataset.

These tests verify that the storage backend integration works correctly while
maintaining backward compatibility with existing code.
"""

import pytest
import pandas as pd
import numpy as np
import tempfile
from pathlib import Path
from datetime import datetime, timedelta

from meteaudata.types import TimeSeries, Signal, Dataset
from meteaudata.storage import StorageConfig, create_backend


@pytest.fixture
def sample_series():
    """Create a sample pandas Series for testing."""
    dates = pd.date_range('2024-01-01', periods=100, freq='h')
    data = np.random.randn(100)
    return pd.Series(data, index=dates, name='test_series')


@pytest.fixture
def sample_signal(sample_series):
    """Create a sample Signal for testing."""
    ts = TimeSeries(series=sample_series.copy())
    signal = Signal(
        name='test_signal',
        time_series={'raw': ts},
        created_on=datetime.now(),
        last_updated=datetime.now()
    )
    return signal


@pytest.fixture
def sample_dataset():
    """Create a sample Dataset for testing."""
    # Create fresh signal to avoid #1 numbering
    dates = pd.date_range('2024-01-01', periods=100, freq='h')
    data = np.random.randn(100)
    series = pd.Series(data, index=dates, name='test_series')

    ts = TimeSeries(series=series)
    signal = Signal(
        name='signal1#1',
        time_series={'raw': ts},
        created_on=datetime.now(),
        last_updated=datetime.now()
    )

    dataset = Dataset(
        name='test_dataset',
        signals={'signal1#1': signal},
        created_on=datetime.now(),
        last_updated=datetime.now()
    )
    return dataset


@pytest.fixture
def temp_dir():
    """Create a temporary directory for file-based backends."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


class TestTimeSeriesBackend:
    """Test backend support for TimeSeries class."""

    def test_timeseries_save_to_backend(self, sample_series, temp_dir):
        """Test TimeSeries can save to backend."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        ts = TimeSeries(series=sample_series)
        ts._backend = backend

        ts.save_to_backend('test_key')

        assert backend.exists('test_key')
        assert ts._storage_key == 'test_key'

    def test_timeseries_load_from_backend(self, sample_series, temp_dir):
        """Test TimeSeries can load from backend."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        # Save first
        ts1 = TimeSeries(series=sample_series)
        ts1._backend = backend
        ts1.save_to_backend('test_key')

        # Load into new TimeSeries
        ts2 = TimeSeries(series=pd.Series([]))
        ts2._backend = backend
        ts2.load_from_backend('test_key')

        pd.testing.assert_series_equal(ts1.series, ts2.series)
        assert ts2._storage_key == 'test_key'

    def test_timeseries_save_without_backend_raises(self, sample_series):
        """Test that saving without backend raises error."""
        ts = TimeSeries(series=sample_series)

        with pytest.raises(ValueError, match="No storage backend configured"):
            ts.save_to_backend('test_key')

    def test_timeseries_load_without_backend_raises(self, sample_series):
        """Test that loading without backend raises error."""
        ts = TimeSeries(series=sample_series)

        with pytest.raises(ValueError, match="No storage backend configured"):
            ts.load_from_backend('test_key')


class TestSignalBackend:
    """Test backend support for Signal class."""

    def test_signal_save_all(self, sample_signal, temp_dir):
        """Test Signal can save all time series to backend."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_signal._backend = backend
        sample_signal.save_all('test_dataset')

        # Keys get sanitized: slashes replaced with underscores
        # Time series keys now include signal prefix, but storage key extracts just the ts part
        expected_key = 'test_dataset_test_signal#1_raw#1'
        assert backend.exists(expected_key)

    def test_signal_save_all_without_backend_skips(self, sample_signal):
        """Test that save_all without backend does nothing."""
        # Should not raise, just skip
        sample_signal.save_all('test_dataset')

    def test_signal_propagates_backend_to_timeseries(self, sample_signal, temp_dir):
        """Test that backend is propagated to time series during save."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_signal._backend = backend
        sample_signal.save_all('test_dataset')

        # Check that time series got backend
        for ts in sample_signal.time_series.values():
            assert ts._backend == backend

    def test_signal_backend_conflict_warning(self, sample_signal, temp_dir):
        """Test that conflicting backends trigger warning."""
        config1 = StorageConfig.for_pandas_disk(temp_dir / "backend1")
        backend1 = create_backend(config1)

        config2 = StorageConfig.for_pandas_disk(temp_dir / "backend2")
        backend2 = create_backend(config2)

        # Get the actual time series name
        ts_names = list(sample_signal.time_series.keys())
        ts_name = ts_names[0]

        # Set different backend on time series
        sample_signal.time_series[ts_name]._backend = backend1

        # Now set different backend on signal
        sample_signal._backend = backend2

        # Should warn when saving
        with pytest.warns(UserWarning, match="already has a different backend"):
            sample_signal.save_all('test_dataset')


class TestDatasetBackend:
    """Test backend support for Dataset class."""

    def test_dataset_set_backend(self, sample_dataset, temp_dir):
        """Test Dataset.set_backend configures backend."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        result = sample_dataset.set_backend(backend, auto_save=True)

        assert sample_dataset._backend == backend
        assert sample_dataset._auto_save is True
        assert result is sample_dataset  # Returns self

    def test_dataset_propagates_backend_to_signals(self, sample_dataset, temp_dir):
        """Test that set_backend propagates to all signals."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_dataset.set_backend(backend, auto_save=True)

        for signal in sample_dataset.signals.values():
            assert signal._backend == backend
            assert signal._auto_save is True
            assert signal._parent_dataset_name == sample_dataset.name

    def test_dataset_save_all(self, sample_dataset, temp_dir):
        """Test Dataset.save_all saves all time series."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_dataset.set_backend(backend)
        sample_dataset.save_all()

        # Check that data was saved (keys get sanitized, ts part is extracted)
        expected_key = 'test_dataset_signal1#1_raw#1'
        assert backend.exists(expected_key)

    def test_dataset_load_all(self, sample_dataset, temp_dir):
        """Test Dataset.load_all loads all time series."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        # Save first
        sample_dataset.set_backend(backend)

        # Get the actual time series name
        signal = sample_dataset.signals['signal1#1']
        ts_names = list(signal.time_series.keys())
        raw_ts_name = ts_names[0]  # Should be 'signal1#1_raw#1'

        original_series = signal.time_series[raw_ts_name].series.copy()
        sample_dataset.save_all()

        # Modify data
        signal.time_series[raw_ts_name].series = pd.Series([1, 2, 3])

        # Load back
        sample_dataset.load_all()

        loaded_series = signal.time_series[raw_ts_name].series
        pd.testing.assert_series_equal(loaded_series, original_series)

    def test_dataset_save_without_backend_raises(self, sample_dataset):
        """Test that save_all without backend raises error."""
        with pytest.raises(ValueError, match="No storage backend configured"):
            sample_dataset.save_all()

    def test_dataset_load_without_backend_raises(self, sample_dataset):
        """Test that load_all without backend raises error."""
        with pytest.raises(ValueError, match="No storage backend configured"):
            sample_dataset.load_all()

    def test_dataset_backend_conflict_warning(self, sample_dataset, temp_dir):
        """Test that conflicting backends trigger warning."""
        config1 = StorageConfig.for_pandas_disk(temp_dir / "backend1")
        backend1 = create_backend(config1)

        config2 = StorageConfig.for_pandas_disk(temp_dir / "backend2")
        backend2 = create_backend(config2)

        # Set different backend on signal
        sample_dataset.signals['signal1#1']._backend = backend1

        # Now set different backend on dataset
        with pytest.warns(UserWarning, match="already has a different backend configured"):
            sample_dataset.set_backend(backend2)


class TestProcessingWithBackend:
    """Test that processing works correctly with backends."""

    def test_signal_process_with_backend_no_autosave(self, sample_signal, temp_dir):
        """Test Signal.process works with backend but no auto-save."""
        from meteaudata.processing_steps.univariate.resample import resample

        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_signal._backend = backend
        sample_signal._auto_save = False

        # Get the full time series name (now includes signal prefix)
        ts_names = list(sample_signal.time_series.keys())
        assert len(ts_names) == 1
        input_ts_name = ts_names[0]  # Should be 'test_signal#1_raw#1'

        # Process should work normally
        sample_signal.process([input_ts_name], resample, frequency='1D')

        # Should have new time series with RESAMPLED suffix (not raw#2)
        expected_new_ts = f'{sample_signal.name}_RESAMPLED#1'
        assert expected_new_ts in sample_signal.time_series

        # But it shouldn't be saved automatically (key sanitized)
        # Note: default dataset name is used since _parent_dataset_name is None
        assert not backend.exists('default_test_signal#1_RESAMPLED#1')

    def test_signal_process_with_auto_save(self, sample_signal, temp_dir):
        """Test Signal.process auto-saves when configured."""
        from meteaudata.processing_steps.univariate.resample import resample

        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_signal._backend = backend
        sample_signal._auto_save = True
        sample_signal._parent_dataset_name = 'test_dataset'

        # Get the full time series name (now includes signal prefix)
        ts_names = list(sample_signal.time_series.keys())
        input_ts_name = ts_names[0]  # Should be 'test_signal#1_raw#1'

        # Process with auto-save
        sample_signal.process([input_ts_name], resample, frequency='1D')

        # Should be saved automatically (key sanitized)
        # save_all saves all time series including the new RESAMPLED#1
        assert backend.exists('test_dataset_test_signal#1_raw#1')
        assert backend.exists('test_dataset_test_signal#1_RESAMPLED#1')

    def test_dataset_process_with_auto_save(self, sample_dataset, temp_dir):
        """Test that signals in dataset auto-save when configured."""
        from meteaudata.processing_steps.univariate.resample import resample

        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_dataset.set_backend(backend, auto_save=True)

        # Get the actual time series name from the signal
        signal = sample_dataset.signals['signal1#1']
        ts_names = list(signal.time_series.keys())
        raw_ts_name = ts_names[0]  # Should be 'signal1#1_raw#1'

        # Process using Signal.process (Dataset.process is for multivariate functions)
        signal.process(
            [raw_ts_name],
            resample,
            frequency='1D'
        )

        # Should be saved automatically (key sanitized)
        assert backend.exists('test_dataset_signal1#1_raw#1')
        assert backend.exists('test_dataset_signal1#1_RESAMPLED#1')


class TestBackwardCompatibility:
    """Test that existing code works without backends."""

    def test_timeseries_without_backend(self, sample_series):
        """Test TimeSeries works normally without backend."""
        ts = TimeSeries(series=sample_series)

        # Should have default None backend
        assert ts._backend is None
        assert ts._storage_key is None

        # Should work normally
        assert len(ts.series) == 100
        assert ts.series.name == 'test_series'

    def test_signal_without_backend(self, sample_signal):
        """Test Signal works normally without backend."""
        # Should have default None backend
        assert sample_signal._backend is None
        assert sample_signal._auto_save is False

        # Should work normally - time series keys now include signal prefix
        ts_names = list(sample_signal.time_series.keys())
        assert len(ts_names) == 1
        ts_name = ts_names[0]
        assert ts_name.endswith('_raw#1')  # Should be 'test_signal#1_raw#1'
        assert len(sample_signal.time_series[ts_name].series) == 100

    def test_dataset_without_backend(self, sample_dataset):
        """Test Dataset works normally without backend."""
        # Should have default None backend
        assert sample_dataset._backend is None
        assert sample_dataset._auto_save is False

        # Should work normally
        assert 'signal1#1' in sample_dataset.signals
        assert len(sample_dataset.signals['signal1#1'].time_series) > 0

    def test_processing_without_backend(self, sample_signal):
        """Test processing works without backend."""
        from meteaudata.processing_steps.univariate.resample import resample

        # Get the input time series name (now includes signal prefix)
        ts_names = list(sample_signal.time_series.keys())
        input_ts_name = ts_names[0]

        # Process without backend
        sample_signal.process([input_ts_name], resample, frequency='1D')

        # Should work normally - should have new time series with RESAMPLED suffix
        expected_new_ts = f'{sample_signal.name}_RESAMPLED#1'
        assert expected_new_ts in sample_signal.time_series


class TestFullWorkflow:
    """Integration tests for complete workflows."""

    def test_full_workflow_with_disk_backend(self, sample_dataset, temp_dir):
        """Test complete workflow with on-disk storage."""
        from meteaudata.processing_steps.univariate.resample import resample

        # Configure backend
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)
        sample_dataset.set_backend(backend, auto_save=False)

        # Get the actual time series name from the signal
        signal = sample_dataset.signals['signal1#1']
        ts_names = list(signal.time_series.keys())
        raw_ts_name = ts_names[0]  # Should be 'signal1#1_raw#1'

        # Process data using Signal.process (Dataset.process is for multivariate functions)
        signal.process([raw_ts_name], resample, frequency='1D')

        # Explicitly save
        sample_dataset.save_all()

        # Verify saved (keys sanitized, and ts part is extracted)
        assert backend.exists('test_dataset_signal1#1_raw#1')
        assert backend.exists('test_dataset_signal1#1_RESAMPLED#1')

        # Modify in memory
        sample_dataset.signals['signal1#1'].time_series[raw_ts_name].series = pd.Series([])

        # Load back
        sample_dataset.load_all()

        # Verify restored
        assert len(sample_dataset.signals['signal1#1'].time_series[raw_ts_name].series) > 0

    def test_workflow_with_sql_backend(self, sample_dataset):
        """Test workflow with SQL backend."""
        from meteaudata.processing_steps.univariate.resample import resample

        # Use in-memory SQLite
        config = StorageConfig.for_sql('sqlite:///:memory:')
        backend = create_backend(config)
        sample_dataset.set_backend(backend, auto_save=True)

        # Get the actual time series name from the signal
        signal = sample_dataset.signals['signal1#1']
        ts_names = list(signal.time_series.keys())
        raw_ts_name = ts_names[0]  # Should be 'signal1#1_raw#1'

        # Process with auto-save using Signal.process (Dataset.process is for multivariate functions)
        signal.process([raw_ts_name], resample, frequency='1D')

        # Should be saved automatically in SQL (keys not sanitized in SQL adapter)
        # Keys are dataset/signal/tspart (where tspart is extracted from full ts name)
        assert backend.exists('test_dataset/signal1#1/raw#1')
        assert backend.exists('test_dataset/signal1#1/RESAMPLED#1')

        # Can query keys
        keys = backend.list_keys()
        assert 'test_dataset/signal1#1/raw#1' in keys

    def test_cross_backend_compatibility(self, sample_dataset, temp_dir):
        """Test moving data between different backends."""
        import warnings

        # Start with pandas-disk
        config1 = StorageConfig.for_pandas_disk(temp_dir / "disk")
        backend1 = create_backend(config1)
        sample_dataset.set_backend(backend1)
        sample_dataset.save_all()

        # Get the actual time series name
        signal = sample_dataset.signals['signal1#1']
        ts_names = list(signal.time_series.keys())
        raw_ts_name = ts_names[0]  # Should be 'signal1#1_raw#1'
        original_series = signal.time_series[raw_ts_name].series.copy()

        # Switch to SQL backend (expect warning about backend change)
        config2 = StorageConfig.for_sql('sqlite:///:memory:')
        backend2 = create_backend(config2)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            sample_dataset.set_backend(backend2)

        # Save to new backend
        sample_dataset.save_all()

        # Clear in-memory data
        sample_dataset.signals['signal1#1'].time_series[raw_ts_name].series = pd.Series([])

        # Load from SQL backend
        sample_dataset.load_all()

        # Should match original
        loaded_series = sample_dataset.signals['signal1#1'].time_series[raw_ts_name].series
        pd.testing.assert_series_equal(loaded_series, original_series)

    def test_multiple_signals_and_timeseries(self, sample_series, temp_dir):
        """Test backend with multiple signals and time series."""
        # Create dataset with multiple signals
        ts1 = TimeSeries(series=sample_series.copy())
        ts2 = TimeSeries(series=sample_series.copy() * 2)

        signal1 = Signal(
            name='signal1#1',
            time_series={'raw': ts1, 'processed': ts2},
            created_on=datetime.now(),
            last_updated=datetime.now()
        )

        signal2 = Signal(
            name='signal2',
            time_series={'raw': ts1.model_copy(deep=True)},
            created_on=datetime.now(),
            last_updated=datetime.now()
        )

        dataset = Dataset(
            name='multi_signal_dataset',
            signals={'signal1#1': signal1, 'signal2': signal2},
            created_on=datetime.now(),
            last_updated=datetime.now()
        )

        # Configure backend
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)
        dataset.set_backend(backend)

        # Save all
        dataset.save_all()

        # Verify all keys exist (keys sanitized)
        # Note: signal1 already has #1, signal2 gets #1 automatically
        # Time series dict keys now include signal prefix, but storage keys extract just the ts part
        expected_keys = [
            'multi_signal_dataset_signal1#1_raw#1',
            'multi_signal_dataset_signal1#1_processed#1',
            'multi_signal_dataset_signal2#1_raw#1'
        ]

        for key in expected_keys:
            assert backend.exists(key)


class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_empty_dataset(self, temp_dir):
        """Test backend with empty dataset."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        dataset = Dataset(
            name='empty_dataset',
            signals={},
            created_on=datetime.now(),
            last_updated=datetime.now()
        )

        dataset.set_backend(backend)
        dataset.save_all()  # Should not raise
        dataset.load_all()  # Should not raise

    def test_load_nonexistent_key(self, sample_dataset, temp_dir):
        """Test loading with missing keys in storage."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_dataset.set_backend(backend)

        # Try to load without saving first
        # This should skip missing keys gracefully
        sample_dataset.load_all()

    def test_hierarchical_key_structure(self, sample_dataset, temp_dir):
        """Test that hierarchical key structure is sanitized."""
        config = StorageConfig.for_pandas_disk(temp_dir)
        backend = create_backend(config)

        sample_dataset.set_backend(backend)
        sample_dataset.save_all()

        keys = backend.list_keys()
        assert len(keys) == 1
        # Keys get sanitized: slashes replaced with underscores
        # Key structure is dataset/signal/tspart where tspart is extracted from full ts name
        assert keys[0] == 'test_dataset_signal1#1_raw#1'


class TestConvenienceMethods:
    """Test the new convenience methods for configuring storage backends."""

    def test_timeseries_use_disk_storage(self, sample_series, temp_dir):
        """Test TimeSeries.use_disk_storage convenience method."""
        ts = TimeSeries(series=sample_series.copy())

        # Use convenience method
        result = ts.use_disk_storage(temp_dir / "ts_data")

        # Should return self for chaining
        assert result is ts

        # Backend should be configured
        assert ts._backend is not None
        assert ts._backend.get_backend_type() == "pandas-disk"

    def test_timeseries_use_sql_storage(self, sample_series):
        """Test TimeSeries.use_sql_storage convenience method."""
        ts = TimeSeries(series=sample_series.copy())

        # Use convenience method with in-memory SQLite
        result = ts.use_sql_storage("sqlite:///:memory:")

        # Should return self for chaining
        assert result is ts

        # Backend should be configured
        assert ts._backend is not None
        assert ts._backend.get_backend_type() == "sql"

    def test_timeseries_use_memory_storage(self, sample_series):
        """Test TimeSeries.use_memory_storage convenience method."""
        ts = TimeSeries(series=sample_series.copy())

        # Use convenience method
        result = ts.use_memory_storage()

        # Should return self for chaining
        assert result is ts

        # Backend should be configured
        assert ts._backend is not None
        assert ts._backend.get_backend_type() == "pandas-memory"

    def test_signal_use_disk_storage(self, sample_signal, temp_dir):
        """Test Signal.use_disk_storage convenience method."""
        # Use convenience method
        result = sample_signal.use_disk_storage(temp_dir / "signal_data", auto_save=True)

        # Should return self for chaining
        assert result is sample_signal

        # Backend should be configured
        assert sample_signal._backend is not None
        assert sample_signal._backend.get_backend_type() == "pandas-disk"
        assert sample_signal._auto_save is True

        # Backend should be propagated to time series
        for ts in sample_signal.time_series.values():
            assert ts._backend is sample_signal._backend

    def test_signal_use_sql_storage(self, sample_signal):
        """Test Signal.use_sql_storage convenience method."""
        # Use convenience method
        result = sample_signal.use_sql_storage("sqlite:///:memory:", auto_save=False)

        # Should return self for chaining
        assert result is sample_signal

        # Backend should be configured
        assert sample_signal._backend is not None
        assert sample_signal._backend.get_backend_type() == "sql"
        assert sample_signal._auto_save is False

        # Backend should be propagated to time series
        for ts in sample_signal.time_series.values():
            assert ts._backend is sample_signal._backend

    def test_dataset_use_disk_storage(self, sample_dataset, temp_dir):
        """Test Dataset.use_disk_storage convenience method."""
        # Use convenience method
        result = sample_dataset.use_disk_storage(temp_dir / "dataset_data")

        # Should return self for chaining
        assert result is sample_dataset

        # Backend should be configured
        assert sample_dataset._backend is not None
        assert sample_dataset._backend.get_backend_type() == "pandas-disk"
        assert sample_dataset._auto_save is True  # Default is True

        # Backend should be propagated to signals and time series
        for signal in sample_dataset.signals.values():
            assert signal._backend is sample_dataset._backend
            assert signal._auto_save is True
            for ts in signal.time_series.values():
                assert ts._backend is sample_dataset._backend

    def test_dataset_use_sql_storage(self, sample_dataset):
        """Test Dataset.use_sql_storage convenience method."""
        # Use convenience method
        result = sample_dataset.use_sql_storage("sqlite:///:memory:", auto_save=False)

        # Should return self for chaining
        assert result is sample_dataset

        # Backend should be configured
        assert sample_dataset._backend is not None
        assert sample_dataset._backend.get_backend_type() == "sql"
        assert sample_dataset._auto_save is False

        # Backend should be propagated to signals and time series
        for signal in sample_dataset.signals.values():
            assert signal._backend is sample_dataset._backend
            assert signal._auto_save is False
            for ts in signal.time_series.values():
                assert ts._backend is sample_dataset._backend

    def test_dataset_use_memory_storage(self, sample_dataset):
        """Test Dataset.use_memory_storage convenience method."""
        # Use convenience method
        result = sample_dataset.use_memory_storage()

        # Should return self for chaining
        assert result is sample_dataset

        # Backend should be configured
        assert sample_dataset._backend is not None
        assert sample_dataset._backend.get_backend_type() == "pandas-memory"
        assert sample_dataset._auto_save is False  # Memory doesn't need auto-save

        # Backend should be propagated to signals and time series
        for signal in sample_dataset.signals.values():
            assert signal._backend is sample_dataset._backend
            for ts in signal.time_series.values():
                assert ts._backend is sample_dataset._backend

    def test_convenience_methods_with_pathlib(self, sample_dataset, temp_dir):
        """Test that convenience methods work with pathlib.Path objects."""
        from pathlib import Path

        # Should work with Path object
        path = Path(temp_dir) / "test_path"
        result = sample_dataset.use_disk_storage(path)

        assert result is sample_dataset
        assert sample_dataset._backend is not None

    def test_convenience_methods_chaining(self, sample_dataset, temp_dir):
        """Test that convenience methods can be chained."""
        # Should be able to chain method calls
        result = sample_dataset.use_disk_storage(temp_dir / "data")

        # Can continue chaining other dataset methods
        signal = result.signals['signal1#1']
        assert signal is not None



"""Polars storage adapter.

This adapter uses Polars for data storage and processing, providing
better performance for certain operations while maintaining compatibility
with the pandas-based system through conversion utilities.
"""

from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional, Union, TYPE_CHECKING
import pandas as pd
import yaml
import tempfile
import shutil

from meteaudata.types import IndexMetadata

try:
    import polars as pl
    POLARS_AVAILABLE = True
except ImportError:
    POLARS_AVAILABLE = False
    pl = None  # type: ignore

if TYPE_CHECKING:
    import polars as pl


class PolarsAdapter:
    """Storage adapter using Polars for data storage.

    This adapter stores data using Polars Series and Parquet files,
    providing better performance for certain operations. It automatically
    converts between pandas and Polars as needed.

    Attributes:
        base_path: Base directory for storing data files

    Raises:
        ImportError: If Polars is not installed
    """

    def __init__(self, base_path: Optional[Path] = None):
        """Initialize the Polars storage adapter.

        Args:
            base_path: Directory to store files. If None, creates a temporary directory.

        Raises:
            ImportError: If Polars is not installed
        """
        if not POLARS_AVAILABLE:
            raise ImportError(
                "Polars is required for PolarsAdapter. "
                "Install it with: pip install polars"
            )

        if base_path is None:
            base_path = Path(tempfile.mkdtemp(prefix="meteaudata_polars_"))

        self._base_path = Path(base_path)
        self._base_path.mkdir(parents=True, exist_ok=True)

    @property
    def base_path(self) -> Path:
        """Get the base path for file storage.

        Returns:
            Path to the base directory
        """
        return self._base_path

    def get_data_path(self, key: str) -> Path:
        """Get the file path for storing data.

        Args:
            key: Unique identifier

        Returns:
            Path to the Parquet data file
        """
        safe_key = key.replace("/", "_").replace("\\", "_")
        return self._base_path / f"{safe_key}.parquet"

    def get_metadata_path(self, key: str) -> Path:
        """Get the file path for storing metadata.

        Args:
            key: Unique identifier

        Returns:
            Path to the YAML metadata file
        """
        safe_key = key.replace("/", "_").replace("\\", "_")
        return self._base_path / f"{safe_key}_meta.yaml"

    def save(self, data: Union[pl.Series, pd.Series], key: str, metadata: Dict[str, Any]) -> None:
        """Save data and metadata to disk.

        Args:
            data: Polars or Pandas Series to store
            key: Unique identifier
            metadata: Metadata dictionary
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        # Convert pandas to Polars if needed
        if isinstance(data, pd.Series):
            # Extract and store index metadata before conversion
            index_metadata = IndexMetadata.extract_index_metadata(data.index)
            metadata['index_metadata'] = index_metadata.model_dump()

            # Store the actual index values for reconstruction
            # Convert to string for YAML serialization (works for all index types)
            metadata['index_values'] = [str(val) for val in data.index]

            # Store series name in metadata before conversion
            if 'original_name' not in metadata and data.name is not None:
                metadata['original_name'] = data.name

            # Convert to Polars
            pl_series = pl.from_pandas(data)
        else:
            pl_series = data

        # Convert Series to DataFrame for Parquet storage
        series_name = pl_series.name or metadata.get('original_name', 'value')
        df = pl.DataFrame({series_name: pl_series})

        # Save data as Parquet
        df.write_parquet(data_path, compression='snappy')

        # Save metadata as YAML
        with open(meta_path, 'w') as f:
            yaml.dump(metadata, f, default_flow_style=False, allow_unicode=True)

    def load(self, key: str) -> Tuple[pl.Series, Dict[str, Any]]:
        """Load data and metadata from disk.

        Args:
            key: Unique identifier

        Returns:
            Tuple of (Polars series, metadata)

        Raises:
            KeyError: If key does not exist
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        if not data_path.exists():
            raise KeyError(f"Key '{key}' not found in storage (path: {data_path})")

        # Load data from Parquet
        df = pl.read_parquet(data_path)

        # Convert to Series (take first column)
        series = df.to_series(0)

        # Load metadata from YAML
        if meta_path.exists():
            with open(meta_path, 'r') as f:
                metadata = yaml.safe_load(f) or {}
        else:
            metadata = {}

        # Note: Index values and metadata are stored separately and will be
        # reconstructed when converting to pandas via to_pandas()
        # Polars doesn't have the same index concept, so we store it in metadata

        return series, metadata

    def delete(self, key: str) -> None:
        """Delete data and metadata files from disk.

        Args:
            key: Unique identifier

        Raises:
            KeyError: If key does not exist
        """
        data_path = self.get_data_path(key)
        meta_path = self.get_metadata_path(key)

        if not data_path.exists():
            raise KeyError(f"Key '{key}' not found in storage")

        # Delete both data and metadata files
        data_path.unlink()
        if meta_path.exists():
            meta_path.unlink()

    def exists(self, key: str) -> bool:
        """Check if key exists in storage.

        Args:
            key: Unique identifier

        Returns:
            True if data file exists, False otherwise
        """
        return self.get_data_path(key).exists()

    def list_keys(self) -> List[str]:
        """List all stored keys.

        Returns:
            List of all keys in storage
        """
        keys = []
        for parquet_file in self._base_path.glob("*.parquet"):
            # Extract key from filename (remove .parquet extension)
            key = parquet_file.stem
            keys.append(key)
        return sorted(keys)

    def to_pandas(self, data: Union[pl.Series, pd.Series]) -> pd.Series:
        """Convert to pandas Series.

        Note: This method only handles the data conversion. Index metadata
        reconstruction must be done separately after loading, as the metadata
        is not part of the Polars series itself.

        Args:
            data: Polars or Pandas Series

        Returns:
            Pandas Series
        """
        if isinstance(data, pd.Series):
            return data
        return data.to_pandas()

    def from_pandas(self, series: pd.Series) -> pl.Series:
        """Convert from pandas Series to Polars Series.

        Args:
            series: Pandas Series

        Returns:
            Polars Series
        """
        return pl.from_pandas(series)

    def clear(self) -> None:
        """Clear all stored data from disk.

        Warning:
            This operation is irreversible and removes all files.
        """
        for file_path in self._base_path.glob("*.parquet"):
            file_path.unlink()
        for file_path in self._base_path.glob("*.yaml"):
            file_path.unlink()

    def cleanup(self) -> None:
        """Remove the entire storage directory.

        Warning:
            This is a destructive operation that removes the base directory
            and all its contents. Use with caution.
        """
        if self._base_path.exists():
            shutil.rmtree(self._base_path)

    def get_backend_type(self) -> str:
        """Get backend type identifier.

        Returns:
            String 'polars'
        """
        return "polars"

    def get_size_on_disk(self) -> int:
        """Get total size of stored data in bytes.

        Returns:
            Total size in bytes
        """
        total_size = 0
        for file_path in self._base_path.glob("*"):
            if file_path.is_file():
                total_size += file_path.stat().st_size
        return total_size

    def __len__(self) -> int:
        """Get number of stored items.

        Returns:
            Number of keys in storage
        """
        return len(self.list_keys())

    def __repr__(self) -> str:
        """String representation.

        Returns:
            Human-readable representation
        """
        return f"PolarsAdapter(base_path='{self._base_path}', keys={len(self)})"



"""
Display extensions for meteaudata objects.
Refactored to use inheritance with minimal code duplication.
"""

from typing import Any, Dict, Optional
from abc import ABC, abstractmethod
from meteaudata.display_utils import (
    _is_notebook_environment,
    _is_complex_object,
    _format_simple_value
)
from meteaudata.tree_builder import TreeBuilder

# HTML style constants
HTML_STYLE = """
<style>
.meteaudata-display {
    font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', 'Noto Sans', Helvetica, Arial, sans-serif;
    font-size: 14px;
    line-height: 1.5;
    color: #24292f;
    background: #ffffff;
    border: 1px solid #d0d7de;
    border-radius: 6px;
    padding: 16px;
    margin: 8px 0;
}
.meteaudata-header {
    font-weight: 600;
    font-size: 16px;
    margin-bottom: 12px;
    color: #0969da;
}
.meteaudata-attr {
    margin: 4px 0;
    padding: 2px 0;
}
.meteaudata-attr-name {
    font-weight: 600;
    color: #656d76;
    display: inline-block;
    min-width: 120px;
}
.meteaudata-attr-value {
    color: #24292f;
}
.meteaudata-nested {
    margin-left: 20px;
    border-left: 2px solid #f6f8fa;
    padding-left: 12px;
    margin-top: 8px;
}
details.meteaudata-collapsible {
    margin: 4px 0;
}
summary.meteaudata-summary {
    cursor: pointer;
    font-weight: 600;
    color: #656d76;
    padding: 4px 0;
}
summary.meteaudata-summary:hover {
    color: #0969da;
}
</style>
"""


class DisplayableBase(ABC):
    """
    Enhanced base class for meteaudata objects with SVG graph visualization.
    """
    
    @abstractmethod
    def _get_display_attributes(self) -> Dict[str, Any]:
        """Get attributes to display. Must be implemented by subclasses."""
        pass
    
    @abstractmethod
    def _get_identifier(self) -> str:
        """Get the key identifier for this object. Must be implemented by subclasses."""
        pass
    
    def __str__(self) -> str:
        """Short description: Object type + key identifier."""
        obj_type = self.__class__.__name__
        identifier = self._get_identifier()
        return f"{obj_type}({identifier})"
    
    def _render_text(self, depth: int, indent: int = 0) -> str:
        """Render text representation."""
        lines = []
        prefix = "  " * indent
        
        # Object header
        lines.append(f"{prefix}{self.__class__.__name__}:")
        
        # Attributes
        for attr_name, attr_value in self._get_display_attributes().items():
            if depth <= 0:
                if hasattr(attr_value, '_render_text'):
                    value_str = str(attr_value)
                else:
                    value_str = f"{type(attr_value).__name__}(...)"
            elif _is_complex_object(attr_value):
                if hasattr(attr_value, '_render_text'):
                    value_str = "\n" + attr_value._render_text(depth - 1, indent + 1)
                else:
                    value_str = str(attr_value)
            else:
                value_str = _format_simple_value(attr_value)
            
            lines.append(f"{prefix}  {attr_name}: {value_str}")
        
        return "\n".join(lines)
    
    def _render_html(self, depth: int) -> None:
        """Render HTML representation with better style injection."""
        try:
            from IPython.display import HTML, display
            
            # Extract CSS content from HTML_STYLE constant
            # Remove <style> and </style> tags and any surrounding whitespace
            css_content = HTML_STYLE.replace('<style>', '').replace('</style>', '').strip()
            
            # Create JavaScript to inject styles
            style_injection = f"""
            <script>
            (function() {{
                var styleId = 'meteaudata-styles';
                if (!document.getElementById(styleId)) {{
                    var style = document.createElement('style');
                    style.id = styleId;
                    style.textContent = `{css_content}`;
                    document.head.appendChild(style);
                }}
            }})();
            </script>
            """
            
            html_content = f"{style_injection}<div class='meteaudata-display'>{self._build_html_content(depth)}</div>"
            display(HTML(html_content))
        except ImportError:
            print(self._render_text(depth))
    
    def _build_html_content(self, depth: int) -> str:
        """Build HTML content for the object using TreeBuilder."""
        # Use TreeBuilder to get the structure
        tree_builder = TreeBuilder()
        tree = tree_builder.build_tree(self, max_depth=depth)
        
        # Render the tree as HTML
        return self._render_tree_node_html(tree, depth)
    
    def _render_tree_node_html(self, node, remaining_depth: int) -> str:
        """Render a TreeNode as HTML."""
        lines = []
        
        # Header
        lines.append(f"<div class='meteaudata-header'>{node.node_type}</div>")
        
        # Simple attributes
        for attr_name, attr_value in node.attributes.items():
            lines.append(f"<div class='meteaudata-attr'><span class='meteaudata-attr-name'>{attr_name}:</span> <span class='meteaudata-attr-value'>{attr_value}</span></div>")
        
        # Child nodes
        if remaining_depth > 0 and node.children:
            for child in node.children:
                if child.is_collection():
                    # Render collection with collapsible details
                    nested_items = []
                    for item_child in child.children:
                        item_content = self._render_tree_node_html(item_child, remaining_depth - 1)
                        nested_items.append(f"<div class='meteaudata-nested'>{item_content}</div>")
                    
                    if nested_items:
                        nested_content = "\n".join(nested_items)
                        lines.append(f"""
                        <details class='meteaudata-collapsible'>
                            <summary class='meteaudata-summary'>{child.name}</summary>
                            <div class='meteaudata-nested'>{nested_content}</div>
                        </details>
                        """)
                else:
                    # Render single object with collapsible details
                    child_content = self._render_tree_node_html(child, remaining_depth - 1)
                    lines.append(f"""
                    <details class='meteaudata-collapsible'>
                        <summary class='meteaudata-summary'>{child.relationship}: {child.node_type}</summary>
                        <div class='meteaudata-nested'>{child_content}</div>
                    </details>
                    """)
        
        return "\n".join(lines)
    
    def render_svg_graph(self, max_depth: int = 4, width: int = 1200, 
                        height: int = 800, title: Optional[str] = None) -> str:
        """
        Render as interactive SVG nested box graph and return HTML string.
        
        Args:
            max_depth: Maximum depth to traverse in object hierarchy
            width: Graph width in pixels
            height: Graph height in pixels
            title: Page title (auto-generated if None)
            
        Returns:
            HTML string with embedded interactive SVG graph
        """
        try:
            from meteaudata.graph_display import SVGNestedBoxGraphRenderer
            
            if title is None:
                title = f"Interactive {self.__class__.__name__} Hierarchy"
            
            renderer = SVGNestedBoxGraphRenderer()
            return renderer.render_to_html(self, max_depth, width, height, title)
        except ImportError:
            raise ImportError(
                "SVG graph rendering requires the svg_nested_boxes module. "
                "Please ensure meteaudata is properly installed."
            )
    
    def show_graph_in_browser(self, max_depth: int = 4, width: int = 1200, 
                             height: int = 800, title: Optional[str] = None) -> str:
        """
        Render SVG graph and open in browser.
        
        Args:
            max_depth: Maximum depth to traverse in object hierarchy
            width: Graph width in pixels
            height: Graph height in pixels
            title: Page title (auto-generated if None)
            
        Returns:
            Path to the generated HTML file
        """
        try:
            from meteaudata.graph_display import open_meteaudata_graph_in_browser
            
            if title is None:
                title = f"Interactive {self.__class__.__name__} Hierarchy"
            
            return open_meteaudata_graph_in_browser(self, max_depth, width, height, title)
        except ImportError:
            raise ImportError(
                "Browser functionality requires additional modules. "
                "Please ensure meteaudata is properly installed."
            )
    
    def save_html(self, path: str, depth: int = 2) -> None:
        """
        Save HTML representation to a file.
        
        Args:
            path: Path to save the HTML file
            depth: Depth of the nested object tree to display
        """
        # Extract CSS content from HTML_STYLE constant
        css_content = HTML_STYLE.replace('<style>', '').replace('</style>', '').strip()
        
        # Create complete HTML document
        html_content = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <meta charset="utf-8">
            <style>
            {css_content}
            body {{ margin: 0; padding: 0; }}
            </style>
        </head>
        <body>
            <div class='meteaudata-display'>
                {self._build_html_content(depth)}
            </div>
        </body>
        </html>
        """
        
        with open(path, 'w', encoding='utf-8') as f:
            f.write(html_content)

    def display(self, format: str = "html", depth: int = 2, 
            max_depth: int = 4, width: int = 1200, height: int = 800,
            output_file: Optional[str] = None) -> None:
        """
        Display method with support for text, HTML, and interactive graph formats.
        
        Args:
            format: Output format ('text', 'html', 'graph')
            depth: Depth for text/html display
            max_depth: Max depth for graph display
            width: Width for graph display
            height: Height for graph display
            output_file: Optional path to save the output (HTML only)
        """
        if format == "text":
            print(self._render_text(depth))
        elif format == "html":
            if output_file:
                self.save_html(output_file, depth)
            else:
                self._render_html(depth)
        elif format == "graph":
            if _is_notebook_environment():
                try:
                    from IPython.display import HTML, display
                    # Check if the imported objects are actually usable (not None)
                    if HTML is None or display is None:
                        raise ImportError("IPython.display components are None")
                    html_content = self.render_svg_graph(max_depth, width, height)
                    display(HTML(html_content))
                except (ImportError, AttributeError, TypeError):
                    print("Notebook environment detected but IPython not available.")
                    print("Use show_graph_in_browser() to open in browser instead.")
            else:
                self.show_graph_in_browser(max_depth, width, height)
        else:
            raise ValueError(f"Unknown format: {format}. Use 'text', 'html', or 'graph'")
    
    # Convenience methods for quick access to different display modes
    def show_details(self, depth: int = 3) -> None:
        """
        Convenience method to show detailed HTML view.
        
        Args:
            depth: How deep to expand nested objects
        """
        self.display(format="html", depth=depth)
    
    def show_summary(self) -> None:
        """
        Convenience method to show a text summary.
        """
        self.display(format="text", depth=1)
    
    def show_graph(self, max_depth: int = 4, width: int = 1200, height: int = 800) -> None:
        """
        Convenience method to show the interactive graph.
        
        Args:
            max_depth: Maximum depth to traverse in object hierarchy
            width: Graph width in pixels  
            height: Graph height in pixels
        """
        self.display(format="graph", max_depth=max_depth, width=width, height=height)


"""Series container with lazy loading support."""

from typing import Any, Optional, Dict, TYPE_CHECKING
import pandas as pd

if TYPE_CHECKING:
    from meteaudata.storage.protocols import StorageBackend, SeriesLike


class SeriesContainer:
    """Container for time series data with lazy loading support.

    This class wraps time series data and provides lazy loading capabilities.
    Data is only loaded from the backend when actually accessed, enabling
    memory-efficient processing of large datasets.

    Attributes:
        _backend: The storage backend to use
        _key: Unique identifier for this data in the backend
        _data: The actual data (None if not yet loaded)
        _loaded: Whether data has been loaded into memory
        _metadata: Associated metadata
    """

    def __init__(
        self,
        backend: "StorageBackend",
        key: Optional[str] = None,
        data: Optional["SeriesLike"] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the series container.

        Args:
            backend: Storage backend to use
            key: Unique identifier (required if data not provided)
            data: Optional pre-loaded data
            metadata: Optional metadata dict
        """
        if data is None and key is None:
            raise ValueError("Either 'data' or 'key' must be provided")

        self._backend = backend
        self._key = key
        self._data = data
        self._metadata = metadata or {}
        self._loaded = data is not None

    @property
    def data(self) -> "SeriesLike":
        """Get the data, loading lazily if needed.

        Returns:
            The time series data

        Raises:
            KeyError: If key doesn't exist in backend
        """
        if not self._loaded:
            if self._key is None:
                raise ValueError("Cannot load data: no key specified")
            self._data, self._metadata = self._backend.load(self._key)
            self._loaded = True
        return self._data

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get the metadata, loading if needed.

        Returns:
            Metadata dictionary
        """
        if not self._metadata and not self._loaded:
            # Trigger load to get metadata
            _ = self.data
        return self._metadata

    @property
    def is_loaded(self) -> bool:
        """Check if data is currently loaded in memory.

        Returns:
            True if data is in memory, False if it needs to be loaded
        """
        return self._loaded

    @property
    def key(self) -> Optional[str]:
        """Get the storage key.

        Returns:
            The unique identifier for this data
        """
        return self._key

    def materialize(self) -> None:
        """Force load data into memory.

        This is useful when you know you'll need the data and want to
        trigger loading explicitly.
        """
        _ = self.data

    def save(self, key: Optional[str] = None, metadata: Optional[Dict[str, Any]] = None) -> None:
        """Save the data to the backend.

        Args:
            key: Optional new key (uses existing if not provided)
            metadata: Optional metadata to save (uses existing if not provided)

        Raises:
            ValueError: If no key is available
            RuntimeError: If data is not loaded
        """
        if not self._loaded:
            raise RuntimeError("Cannot save unloaded data")

        save_key = key or self._key
        if save_key is None:
            raise ValueError("No key specified for saving")

        save_metadata = metadata or self._metadata

        self._backend.save(self._data, save_key, save_metadata)
        self._key = save_key
        self._metadata = save_metadata

    def unload(self) -> None:
        """Unload data from memory.

        This frees up memory by discarding the in-memory data. The data can
        be reloaded later from the backend using the stored key.

        Raises:
            ValueError: If no key is available for reloading
        """
        if self._key is None:
            raise ValueError("Cannot unload data without a key for reloading")

        self._data = None
        self._loaded = False

    def to_pandas(self) -> pd.Series:
        """Get data as a pandas Series.

        Returns:
            Pandas Series representation of the data
        """
        return self._backend.to_pandas(self.data)

    def __repr__(self) -> str:
        """String representation.

        Returns:
            Human-readable representation
        """
        status = "loaded" if self._loaded else "not loaded"
        key_info = f"key='{self._key}'" if self._key else "no key"
        return f"SeriesContainer({key_info}, {status})"



"""SQL database storage adapter.

This adapter stores time series data in a SQL database, enabling
SQL-based querying and filtering while supporting memory-efficient
processing of large datasets.
"""

from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional
import pandas as pd
import json
import tempfile

from meteaudata.types import IndexMetadata

try:
    from sqlalchemy import create_engine, text, Engine
    from sqlalchemy.pool import StaticPool
    SQLALCHEMY_AVAILABLE = True
except ImportError:
    SQLALCHEMY_AVAILABLE = False
    create_engine = None
    text = None
    Engine = None
    StaticPool = None


class SQLAdapter:
    """Storage adapter using SQL database.

    This adapter stores time series data in a SQL database using SQLAlchemy.
    Supports SQLite, PostgreSQL, MySQL, and other SQLAlchemy-compatible databases.

    The adapter creates two tables:
    - time_series_data: Stores the actual time series values
    - time_series_metadata: Stores metadata as JSON

    Attributes:
        engine: SQLAlchemy engine for database operations
        connection_string: Database connection string

    Raises:
        ImportError: If SQLAlchemy is not installed
    """

    def __init__(self, connection_string: Optional[str] = None):
        """Initialize the SQL storage adapter.

        Args:
            connection_string: Database URL (e.g., 'sqlite:///path/to/db.db')
                If None, creates a temporary SQLite database.

        Raises:
            ImportError: If SQLAlchemy is not installed
        """
        if not SQLALCHEMY_AVAILABLE:
            raise ImportError(
                "SQLAlchemy is required for SQLAdapter. "
                "Install it with: pip install sqlalchemy"
            )

        if connection_string is None:
            temp_dir = Path(tempfile.mkdtemp(prefix="meteaudata_sql_"))
            db_path = temp_dir / "meteaudata.db"
            connection_string = f"sqlite:///{db_path}"

        self._connection_string = connection_string

        # For SQLite in-memory databases, use StaticPool to maintain connection
        if connection_string.startswith('sqlite:///:memory:'):
            self._engine = create_engine(
                connection_string,
                connect_args={"check_same_thread": False},
                poolclass=StaticPool
            )
        else:
            self._engine = create_engine(connection_string)

        self._ensure_schema()

    @property
    def connection_string(self) -> str:
        """Get the database connection string.

        Returns:
            Connection string
        """
        return self._connection_string

    @property
    def engine(self) -> Engine:
        """Get the SQLAlchemy engine.

        Returns:
            SQLAlchemy engine
        """
        return self._engine

    def _ensure_schema(self) -> None:
        """Create database tables if they don't exist."""
        with self._engine.connect() as conn:
            # Create time series data table
            conn.execute(text("""
                CREATE TABLE IF NOT EXISTS time_series_data (
                    key TEXT NOT NULL,
                    idx INTEGER NOT NULL,
                    index_value TEXT NOT NULL,
                    value REAL,
                    PRIMARY KEY (key, idx)
                )
            """))

            # Create index for faster lookups
            conn.execute(text("""
                CREATE INDEX IF NOT EXISTS idx_key
                ON time_series_data(key)
            """))

            # Create metadata table
            conn.execute(text("""
                CREATE TABLE IF NOT EXISTS time_series_metadata (
                    key TEXT PRIMARY KEY,
                    metadata_json TEXT NOT NULL,
                    series_name TEXT,
                    length INTEGER,
                    dtype TEXT
                )
            """))

            conn.commit()

    def save(self, data: pd.Series, key: str, metadata: Dict[str, Any]) -> None:
        """Save data and metadata to the database.

        Args:
            data: Pandas Series to store
            key: Unique identifier
            metadata: Metadata dictionary
        """
        # Extract and store index metadata
        index_metadata = IndexMetadata.extract_index_metadata(data.index)
        metadata['index_metadata'] = index_metadata.model_dump()

        with self._engine.connect() as conn:
            # Delete existing data for this key
            conn.execute(
                text("DELETE FROM time_series_data WHERE key = :key"),
                {"key": key}
            )
            conn.execute(
                text("DELETE FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )

            # Prepare data for insertion
            records = []
            for idx, (index_val, value) in enumerate(zip(data.index, data.values)):
                records.append({
                    'key': key,
                    'idx': idx,
                    'index_value': str(index_val),
                    'value': float(value) if pd.notna(value) else None
                })

            # Insert data in batches
            if records:
                # Use pandas for efficient bulk insert
                df = pd.DataFrame(records)
                df.to_sql(
                    'time_series_data',
                    conn,
                    if_exists='append',
                    index=False,
                    method='multi'
                )

            # Insert metadata
            metadata_json = json.dumps(metadata, default=str)
            conn.execute(
                text("""
                    INSERT INTO time_series_metadata
                    (key, metadata_json, series_name, length, dtype)
                    VALUES (:key, :metadata_json, :series_name, :length, :dtype)
                """),
                {
                    "key": key,
                    "metadata_json": metadata_json,
                    "series_name": str(data.name) if data.name is not None else None,
                    "length": len(data),
                    "dtype": str(data.dtype)
                }
            )

            conn.commit()

    def load(self, key: str) -> Tuple[pd.Series, Dict[str, Any]]:
        """Load data and metadata from the database.

        Args:
            key: Unique identifier

        Returns:
            Tuple of (series, metadata)

        Raises:
            KeyError: If key does not exist
        """
        with self._engine.connect() as conn:
            # Check if key exists
            result = conn.execute(
                text("SELECT COUNT(*) FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )
            if result.scalar() == 0:
                raise KeyError(f"Key '{key}' not found in storage")

            # Load metadata
            result = conn.execute(
                text("SELECT metadata_json, series_name FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )
            row = result.fetchone()
            metadata = json.loads(row[0])
            series_name = row[1]

            # Load data
            query = """
                SELECT index_value, value
                FROM time_series_data
                WHERE key = :key
                ORDER BY idx
            """
            df = pd.read_sql(text(query), conn, params={"key": key})

            # Reconstruct series
            if len(df) == 0:
                # Empty series
                series = pd.Series([], dtype=float, name=series_name)
            else:
                series = pd.Series(
                    data=df['value'].values,
                    index=df['index_value'].values,
                    name=series_name
                )

            # Reconstruct index with metadata if available
            if 'index_metadata' in metadata:
                index_meta = IndexMetadata(**metadata['index_metadata'])
                reconstructed_index = IndexMetadata.reconstruct_index(
                    series.index,
                    index_meta
                )
                series.index = reconstructed_index

            return series, metadata

    def delete(self, key: str) -> None:
        """Delete data and metadata from the database.

        Args:
            key: Unique identifier

        Raises:
            KeyError: If key does not exist
        """
        with self._engine.connect() as conn:
            # Check if key exists
            result = conn.execute(
                text("SELECT COUNT(*) FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )
            if result.scalar() == 0:
                raise KeyError(f"Key '{key}' not found in storage")

            # Delete data and metadata
            conn.execute(
                text("DELETE FROM time_series_data WHERE key = :key"),
                {"key": key}
            )
            conn.execute(
                text("DELETE FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )
            conn.commit()

    def exists(self, key: str) -> bool:
        """Check if key exists in storage.

        Args:
            key: Unique identifier

        Returns:
            True if key exists, False otherwise
        """
        with self._engine.connect() as conn:
            result = conn.execute(
                text("SELECT COUNT(*) FROM time_series_metadata WHERE key = :key"),
                {"key": key}
            )
            return result.scalar() > 0

    def list_keys(self) -> List[str]:
        """List all stored keys.

        Returns:
            List of all keys in storage
        """
        with self._engine.connect() as conn:
            result = conn.execute(
                text("SELECT key FROM time_series_metadata ORDER BY key")
            )
            return [row[0] for row in result.fetchall()]

    def to_pandas(self, data: pd.Series) -> pd.Series:
        """Convert to pandas Series (no-op for this adapter).

        Args:
            data: Pandas Series

        Returns:
            The same pandas Series
        """
        return data

    def from_pandas(self, series: pd.Series) -> pd.Series:
        """Convert from pandas Series (no-op for this adapter).

        Args:
            series: Pandas Series

        Returns:
            The same pandas Series
        """
        return series

    def clear(self) -> None:
        """Clear all stored data from the database.

        Warning:
            This operation is irreversible and removes all data.
        """
        with self._engine.connect() as conn:
            conn.execute(text("DELETE FROM time_series_data"))
            conn.execute(text("DELETE FROM time_series_metadata"))
            conn.commit()

    def execute_query(self, query: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """Execute a custom SQL query.

        Args:
            query: SQL query string
            params: Optional query parameters

        Returns:
            Query results
        """
        with self._engine.connect() as conn:
            result = conn.execute(text(query), params or {})
            return result.fetchall()

    def get_table_names(self) -> List[str]:
        """Get list of all tables in the database.

        Returns:
            List of table names
        """
        # This is database-specific; using SQLAlchemy's inspector
        from sqlalchemy import inspect
        inspector = inspect(self._engine)
        return inspector.get_table_names()

    def get_backend_type(self) -> str:
        """Get backend type identifier.

        Returns:
            String 'sql'
        """
        return "sql"

    def get_database_size(self) -> Optional[int]:
        """Get the size of the database in bytes.

        Returns:
            Database size in bytes, or None if not available
        """
        if self._connection_string.startswith('sqlite:///'):
            # Extract file path from connection string
            db_path = self._connection_string.replace('sqlite:///', '')
            path = Path(db_path)
            if path.exists():
                return path.stat().st_size
        return None

    def close(self) -> None:
        """Close the database connection."""
        self._engine.dispose()

    def __len__(self) -> int:
        """Get number of stored items.

        Returns:
            Number of keys in storage
        """
        with self._engine.connect() as conn:
            result = conn.execute(
                text("SELECT COUNT(*) FROM time_series_metadata")
            )
            return result.scalar()

    def __repr__(self) -> str:
        """String representation.

        Returns:
            Human-readable representation
        """
        return f"SQLAdapter(connection='{self._connection_string}', keys={len(self)})"

    def __del__(self):
        """Cleanup on deletion."""
        try:
            self.close()
        except Exception:
            pass  # Ignore errors during cleanup


# Alias so that code written against the new name works too.
# SQLAdapter remains the canonical name for one release cycle.
SQLiteAdapter = SQLAdapter
